import re

from spinta.types.datatype import DataType, Time
from spinta.types.datatype import Date
from spinta.types.datatype import DateTime
from spinta.types.geometry.components import Geometry
from spinta.types.datatype import Integer
from spinta.types.datatype import Number

_time_unit_re = re.compile(r'^\d*[YMQWDHTSLUN]$')
_split_time_unit_pattern = re.compile(r'^(\d+)([YMQWDHTSLUN])$')


def is_time_unit(unit: str) -> bool:
    return _time_unit_re.match(unit) is not None


# Initial regular expressions were borrowed from
# https://stackoverflow.com/a/3573731/475477

_prefix = '''
( Y     # 10^24  (yotta)
| Z     # 10^21  (zetta)
| E     # 10^18  (exa)
| P     # 10^15  (peta)
| T     # 10^12  (tera)
| G     # 10^9   (giga)
| M     # 10^6   (mega)
| k     # 10^3   (kilo)
| h     # 10^2   (hecto)
| da    # 10^1   (deca)
| d     # 10^-1  (deci)
| c     # 10^-2  (centi)
| m     # 10^-3  (milli)
| µ     # 10^-6  (micro)
| n     # 10^-9  (nano)
| p     # 10^-12 (pico)
| f     # 10^-15 (femto)
| a     # 10^-18 (atto)
| z     # 10^-21 (zepto)
| y     # 10^-24 (yocto)
)'''

_unit = '''
# SI Base Units
( m     # metre
| g     # gram
| s     # second
| A     # ampere
| K     # kelvin
| mol   # mole
| cd    # candela

# Named units derived from SI Base Units
| Hz    # hertz
| rad   # radian
| sr    # steradian
| N     # newton
| Pa    # pascal
| J     # joule
| W     # watt
| C     # coulomb
| V     # volt
| F     # farad
| Ω     # ohm
| S     # siemens
| Wb    # weber
| T     # tesla
| H     # henry
| °C    # degree Celsius
| lm    # lumen
| lx    # lux
| Bq    # becquerel
| Gy    # gray
| Sv    # sievert
| kat   # katal

# Non-SI units officially accepted for use with the SI
| min   # minute
| h     # hour
| d     # day
| au    # astronomical unit
| °     # degree
| ′     # arcminute
| ″     # arcsecond
| ha    # hectare
| l     # litre
| L     # litre
| t     # tonne
| Da    # dalton
| eV    # electronvolt
| Np    # neper
| B     # bel
| dB    # decibel

# Non-SI units not officially sanctioned for use with the SI
| Gal   # gal (acceleration)
| u     # unified atomic mass unit
| var   # volt-ampere reactive

# Non-SI time units
| yr    # year
| wk    # week
| mo    # month

# Other Non-SI units
| pc        # parsec
| c₀ | c_0  # natural unit of speed
| ħ         # natural unit of action
| mₑ | m_e  # natural unit of mass
| e         # atomic unit of charge
| a₀ | a_0  # atomic unit of length
| E_h       # atomic unit of energy
| M         # nautical mile
| kn        # knot
| Å         # ångström
| a         # are
| b         # barn
| bar       # bar
| atm       # standard atmosphere
| Ci        # curie
| R         # roentgen
| rem       # rem
| erg       # erg
| dyn       # dyne
| P         # poise
| st        # stokes
| Mx        # maxwell
| G         # gauss
| Oe        # ørsted
| sb        # stilb
| ph        # phot
| Torr      # torr
| kgf       # kilogram-force
| cal       # calorie
| μ         # micron
| xu        # x-unit
| γ         # gamma (mass, magnetic flux density)
| λ         # lambda
| Jy        # jansky
| mmHg      # millimetre of mercury

# Special units
| U     # unit
| %     # percent
)'''
_power = r'([⁺⁻]?[¹²³⁴⁵⁶⁷⁸⁹][⁰¹²³⁴⁵⁶⁷⁸⁹]*|\^[+-]?[1-9]\d*)'
_unit_and_prefix = '((' + _prefix + '?' + _unit + ')' + _power + '?|1)'
_multiplied = _unit_and_prefix + '(?:[⋅·*]' + _unit_and_prefix + ')*'
_with_denominator = _multiplied + '(?:/' + _multiplied + ')?'
_expr = r'\d*' + _power + '?' + _with_denominator
_si_unit_re = re.compile('^' + _expr + r'(?:\ ' + _expr + ')*$', flags=re.VERBOSE)


def is_si_unit(unit: str) -> bool:
    return _si_unit_re.match(unit) is not None


def is_unit(dtype: DataType, unit: str) -> bool:
    if isinstance(dtype, (Date, DateTime, Time)):
        return is_time_unit(unit)
    if isinstance(dtype, (Integer, Number, Geometry)):
        return is_si_unit(unit)
    return False


def split_time_unit(unit: str):
    match = _split_time_unit_pattern.match(unit)
    if match:
        return match.groups()
    return None



import json
import pathlib
from typing import Type

from ruamel.yaml import YAML

from spinta.auth import client_name_exists, get_clients_path, handle_auth_client_files
from spinta.formats.components import Format
from spinta.core.config import DEFAULT_CONFIG_PATH
from spinta.core.config import DEFAULT_DATA_PATH
from spinta.utils.config import asbool
from spinta.utils.imports import importstr
from spinta.commands import load, check
from spinta.components import Context, Config
from spinta import components
from spinta.core.ufuncs import ufunc

yaml = YAML(typ='safe')


@load.register(Context, Config)
def load(context: Context, config: Config) -> Config:
    # Save reference to raw config and give chance for other components to read
    # directly from raw config.
    rc = config.rc = context.get('rc')

    # Load commands.
    config.commands = {}
    for scope in rc.keys('commands'):
        if scope == 'modules':
            continue
        config.commands[scope] = {}
        for name in rc.keys('commands', scope):
            command = rc.get('commands', scope, name, cast=importstr)
            config.commands[scope][name] = command

    # Load ufuncs.
    ufunc.resolver.collect(rc.get('ufuncs'))
    config.resolvers = ufunc.resolver.ufuncs()
    config.executors = ufunc.executor.ufuncs()

    # Load components.
    # TODO: Replace components from dict to classes with type annotations.
    config.components = {}
    for group in rc.keys('components'):
        config.components[group] = {}
        for name in rc.keys('components', group):
            component = rc.get('components', group, name, cast=importstr)
            config.components[group][name] = component

    # Load exporters.
    config.exporters = {}
    for name in rc.keys('exporters'):
        exporter: Type[Format] = rc.get('exporters', name, cast=importstr)
        config.exporters[name] = exporter()

    # Load accesslog
    accesslog = rc.get('accesslog', 'type', required=True)
    config.AccessLog = config.components['accesslog'][accesslog]

    # Load everything else.
    config.debug = rc.get('debug', default=False)
    config.config_path = pathlib.Path(
        rc.get('config_path') or
        DEFAULT_CONFIG_PATH
    )
    config.data_path = pathlib.Path(
        rc.get('data_path') or
        DEFAULT_DATA_PATH,
    )
    config.credentials_file = pathlib.Path(
        rc.get('credentials_file') or
        DEFAULT_CONFIG_PATH / 'credentials.cfg'
    )
    config.server_url = rc.get('server_url')
    config.scope_prefix = rc.get('scope_prefix')
    config.scope_formatter = rc.get('scope_formatter', cast=importstr)
    config.scope_max_length = rc.get('scope_max_length', cast=int)
    config.default_auth_client = rc.get('default_auth_client')
    config.http_basic_auth = rc.get('http_basic_auth', default=False, cast=asbool)
    config.token_validation_key = rc.get('token_validation_key', cast=json.loads) or None
    config.datasets = rc.get('datasets', default={})
    config.env = rc.get('env')
    config.docs_path = rc.get('docs_path', default=None)
    config.always_show_id = rc.get('always_show_id', default=False)
    config.push_page_size = rc.get('push_page_size', default=None, cast=int)
    config.languages = rc.get('languages', default=[])
    config.root = rc.get('root', default=None)
    if config.root is not None:
        config.root = config.root.strip().strip('/')

    # XXX: A check to make sure, that I don't add mode to the config root, which
    #      already happened several times. Remove this, when configuration
    #      ensures that there are no unknown configuration options added.
    if rc.get('mode') is not None:
        raise RuntimeError(
            "Configuration option `mode` must be added to a manifest, now it "
            "is added to the config root."
        )
    # Ensure that client filed were migrated
    handle_auth_client_files(config)

    return config


@check.register(Context, components.Config)
def check(context: Context, config: components.Config):
    path = get_clients_path(config)
    if config.default_auth_client and not client_name_exists(path, config.default_auth_client):
        clients_dir = path
        raise Exception(f"default_auth_client {config.default_auth_client!r} does not exist in {clients_dir}.")



from typing import Iterator
from typing import List
from typing import Optional

from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import Typer
from typer import echo

from spinta.cli.helpers.store import load_manifest
from spinta.components import Context
from spinta.core.context import configure_context
from spinta.core.enums import Access
from spinta.manifests.tabular.components import ManifestColumn
from spinta.manifests.tabular.components import ManifestRow
from spinta.manifests.tabular.helpers import datasets_to_tabular
from spinta.manifests.tabular.helpers import normalizes_columns
from spinta.manifests.tabular.helpers import render_tabular_manifest_rows
from spinta.manifests.tabular.helpers import write_tabular_manifest
from spinta.naming.helpers import reformat_names
from spinta.utils.enums import get_enum_by_name

app = Typer()


@app.command(short_help="Copy manifest optionally transforming final copy")
def copy(
    ctx: TyperContext,
    source: bool = Option(True, help=(
        "Do not copy external data source metadata"
    )),
    # TODO: Change `str` to `Access`
    #       https://github.com/tiangolo/typer/issues/151
    access: str = Option('private', help=(
        "Copy properties with at least specified access"
    )),
    format_names: bool = Option(False, help=(
        "Reformat model and property names."
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output tabular manifest in a specified file"
    )),
    columns: Optional[str] = Option(None, '-c', '--columns', help=(
        "Comma separated list of columns"
    )),
    order_by: Optional[str] = Option(None, help=(
        "Order by a specified column (currently only access column is supported)"
    )),
    rename_duplicates: bool = Option(False, help=(
        "Rename duplicate model names by adding number suffix"
    )),
    manifests: List[str] = Argument(None, help=(
        "Source manifest files to copy from"
    )),
):
    """Copy models from CSV manifest files into another CSV manifest file"""
    context: Context = ctx.obj
    access = get_enum_by_name(Access, access)
    cols = normalizes_columns(columns.split(',')) if columns else None

    verbose = True
    if not output:
        verbose = False

    rows = _read_csv_files(
        context,
        manifests,
        external=source,
        access=access,
        format_names=format_names,
        order_by=order_by,
        rename_duplicates=rename_duplicates,
        verbose=verbose,
    )

    if output:
        write_tabular_manifest(output, rows)
    else:
        echo(render_tabular_manifest_rows(rows, cols))


def _read_csv_files(
    context: Context,
    manifests: List[str],
    *,
    external: bool = True,
    access: Access = Access.private,
    format_names: bool = False,
    order_by: ManifestColumn = None,
    rename_duplicates: bool = False,
    verbose: bool = True,
) -> Iterator[ManifestRow]:
    context = configure_context(context, manifests)
    store = load_manifest(
        context,
        rename_duplicates=rename_duplicates,
        load_internal=False,
        verbose=verbose,
    )

    if format_names:
        reformat_names(context, store.manifest)

    yield from datasets_to_tabular(
        store.manifest,
        external=external,
        access=access,
        order_by=order_by,
    )




from typing import Any
from typing import overload

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import DataType
from spinta.types.datatype import Integer
from spinta.types.datatype import String
from spinta.types.datatype import Boolean
from spinta.dimensions.enum.components import EnumItem
from spinta.exceptions import InvalidValue
from spinta.utils.schema import NotAvailable


@overload
@commands.check.register(Context, EnumItem, DataType, object)
def check(
    context: Context,
    item: EnumItem,
    dtype: DataType,
    value: Any,
):
    raise InvalidValue(dtype, error=(
        f"Given enum value {value} of {type(value)} type does not match "
        f"property type, which is {dtype.name!r}."
    ))


@overload
@commands.check.register(Context, EnumItem, DataType, type(None))
def check(
    context: Context,
    item: EnumItem,
    dtype: DataType,
    value: None,
):
    pass


@overload
@commands.check.register(Context, EnumItem, DataType, NotAvailable)
def check(
    context: Context,
    item: EnumItem,
    dtype: DataType,
    value: NotAvailable,
):
    pass


@overload
@commands.check.register(Context, EnumItem, Integer, int)
def check(
    context: Context,
    item: EnumItem,
    dtype: Integer,
    value: int,
):
    pass


@overload
@commands.check.register(Context, EnumItem, Boolean, bool)
def check(
    context: Context,
    item: EnumItem,
    dtype: Boolean,
    value: bool,
):
    pass


@overload
@commands.check.register(Context, EnumItem, String, str)
def check(
    context: Context,
    item: EnumItem,
    dtype: Integer,
    value: str,
):
    pass



from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

from spinta.manifests.components import Manifest


class YamlManifest(Manifest):
    type = 'yaml'
    path: Path = None


InlineManifestData = List[Dict[str, Any]]


class InlineManifest(YamlManifest):
    type = 'inline'
    manifest: InlineManifestData



from io import TextIOBase
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import Literal
from typing import Optional
from typing import Tuple

import sqlparse
from sqlparse.sql import Identifier
from sqlparse.sql import Statement
from sqlparse.sql import Token
from sqlparse.sql import TokenList
from sqlparse.tokens import DDL
from sqlparse.tokens import Keyword

from spinta import commands
from spinta import spyna
from spinta.components import Context
from spinta.datasets.backends.sqldump.components import SqlDump
from spinta.datasets.backends.sqldump.ufuncs.components import File
from spinta.datasets.backends.sqldump.ufuncs.components import \
    PrepareFileResource
from spinta.datasets.components import Resource
from spinta.exceptions import UnexpectedFormulaResult
from spinta.manifests.components import ManifestSchema
from spinta.utils.naming import to_model_name


class ReadTokens:

    def __init__(self, token_list: TokenList, idx=0):
        self.token_list = token_list
        self.idx = idx

    def find(self, **kwargs) -> Optional[Token]:
        idx, token = self.token_list.token_next_by(idx=self.idx, **kwargs)
        if idx is not None:
            self.idx = idx
            return token


def _read_sql_statement(
    resource: Optional[Resource],
    statement: Statement,
) -> Iterator[Tuple[int, Dict[str, Any]]]:
    tokens = ReadTokens(statement)
    if tokens.find(m=(DDL, 'CREATE')) and tokens.find(m=(Keyword, 'TABLE')):
        identifier: Optional[Identifier] = tokens.find(i=Identifier)
        if identifier is None:
            return
        real_name = identifier.get_real_name()
        name = to_model_name(real_name)
        if resource:
            name = f'{resource.dataset.name}/{name}'
        yield 0, {
            'type': 'model',
            'name': name,
            'external': {
                'dataset': resource.dataset.name if resource else None,
                'resource': resource.name if resource else None,
                'name': real_name,
                'pk': [],
            },
            'properties': {},
        }


def _read_sql_ast(
    resource: Resource,
    ast: Iterable[Statement],
) -> Iterator[Tuple[int, Dict[str, Any]]]:
    for statement in ast:
        yield from _read_sql_statement(resource, statement)


@commands.inspect.register(Context, SqlDump, Resource, type(None))
def inspect(
    context: Context,
    backend: SqlDump,
    resource: Resource,
    source: Literal[None],
) -> Iterator[ManifestSchema]:

    if resource.prepare:
        env = PrepareFileResource(context).init(backend.path)
        file = env.resolve(resource.prepare)
        file: File = env.execute(file)
        f = file.open()
        if not isinstance(f, TextIOBase):
            raise UnexpectedFormulaResult(
                resource,
                formula=spyna.unparse(resource.prepare),
                expected='TextIO',
                received=type(f).__name__,
            )
    else:
        f = backend.stream or backend.path.open()

    with f:
        ast = sqlparse.parsestream(f)
        yield from _read_sql_ast(resource, ast)



import asyncio
import itertools
import json
from typing import Optional

from typer import Context as TyperContext
from typer import Option

from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.store import prepare_manifest
from spinta.cli.helpers.data import process_stream
from spinta.commands.write import write


def import_(
    ctx: TyperContext,
    source: str,
    auth: Optional[str] = Option(None, '-a', '--auth', help=(
        "Authorize as a client"
    )),
    limit: Optional[int] = Option(None, help=(
        "Limit number of rows read from source."
    )),
):
    """Import data from a file"""
    context = ctx.obj
    store = prepare_manifest(context)
    manifest = store.manifest
    root = manifest.objects['ns']['']

    with context:
        require_auth(context, auth)
        context.attach('transaction', manifest.backend.transaction, write=True)
        with open(source) as f:
            stream = (json.loads(line.strip()) for line in f)
            stream = itertools.islice(stream, limit) if limit else stream
            stream = write(context, root, stream, changed=True)
            coro = process_stream(source, stream)
            asyncio.get_event_loop().run_until_complete(coro)



from spinta.commands import command
from spinta.components import Context


@command()
def replace():
    pass


@replace.register()
def replace(context: Context, *, source: dict, data):
    return source.get(data, data)


@command()
def hint():
    pass


@command()
def self():
    pass


@command()
def chain():
    pass


@command('all')
def all_():
    pass


@command()
def denormalize():
    pass


@command()
def unstack():
    pass


@command('list')
def list_():
    pass


@list_.register()
def list_(context: Context, *, commands, **kwargs):
    result = []
    for commands_ in commands:
        for cmd in commands_:
            result.append(cmd(context, **kwargs))
    return result


@command()
def url():
    pass


@command('range')
def range_():
    pass


@range_.register()
def range_(context: Context, *, start: int = 0, stop: int = 0):
    return range(start, stop + 1)



import csv
from typing import Any
from typing import Dict
from typing import Iterator

from starlette.requests import Request
from starlette.responses import Response
from starlette.responses import StreamingResponse

from spinta import commands
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.components import UrlParams
from spinta.formats.csv.components import Csv
from spinta.formats.csv.components import IterableFile
from spinta.formats.helpers import get_model_tabular_header
from spinta.utils.nestedstruct import flatten
from spinta.utils.response import aiter


@commands.render.register(Context, Request, Model, Csv)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Csv,
    *,
    action: Action,
    params: UrlParams,
    data: Iterator[Dict[str, Any]],
    status_code: int = 200,
    headers: Dict[str, str] = None,
) -> Response:
    headers = headers or {}
    headers['Content-Disposition'] = f'attachment; filename="{model.basename}.csv"'
    return StreamingResponse(
        aiter(_render_csv(context, model, action, params, data)),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )


def _render_csv(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
    data: Iterator[Dict[str, Any]],
):
    rows = flatten(data)
    cols = get_model_tabular_header(context, model, action, params)

    stream = IterableFile()
    writer = csv.DictWriter(stream, fieldnames=cols)
    writer.writeheader()
    for row in rows:
        writer.writerow(row)
        yield from stream



from typing import Union, List

from spinta.components import Page, Property, PageBy
from spinta.core.ufuncs import ufunc, Expr, asttoexpr, Negative, Bind, Positive, Pair, Env
from spinta.exceptions import FieldNotInResource, InvalidArgumentInExpression, CannotSelectTextAndSpecifiedLang
from spinta.types.datatype import DataType, String
from spinta.types.text.components import Text
from spinta.ufuncs.basequerybuilder.components import BaseQueryBuilder, LoadBuilder
from spinta.ufuncs.helpers import merge_formulas


@ufunc.resolver(BaseQueryBuilder, Expr, name='paginate')
def paginate(env, expr):
    if len(expr.args) != 1:
        raise InvalidArgumentInExpression(arguments=expr.args, expr='paginate')
    page = expr.args[0]
    if isinstance(page, Page):
        if page.is_enabled:
            if not page.filter_only:
                env.page.select = env.call('select', page)
                for by, page_by in page.by.items():
                    sorted_ = env.call('sort', Negative(page_by.prop.name) if by.startswith("-") else Bind(page_by.prop.name))
                    if sorted_ is not None:
                        env.page.sort.append(sorted_)
            env.page.page_ = page
            env.page.size = page.size
            return env.resolve(_get_pagination_compare_query(page))
    else:
        raise InvalidArgumentInExpression(arguments=expr.args, expr='paginate')


@ufunc.resolver(BaseQueryBuilder, Expr, name='expand')
def expand(env, expr):
    env.expand = []
    if expr.args:
        for arg in expr.args:
            resolved = env.resolve(arg)
            selected = env.call('select', resolved)
            env.expand.append(selected.prop)


@ufunc.resolver(BaseQueryBuilder, Expr, name='page')
def page_(env, expr):
    pass


@ufunc.resolver(LoadBuilder, Expr, name='page')
def page_(env, expr):
    args = env.resolve(expr.args)
    page = Page()
    if len(args) > 0:
        for item in args:
            res = env.resolve(item)
            res = env.call('page_item', res)
            if isinstance(res[1], Property):
                page.by[res[0]] = PageBy(res[1])
            else:
                if res[0] == 'size' and isinstance(res[1], int):
                    page.size = res[1]
                else:
                    raise InvalidArgumentInExpression(arguments=res[0], expr='page')
    else:
        page.is_enabled = False
    return page


@ufunc.resolver(LoadBuilder, Pair, name='page_item')
def page_item(env, field):
    return field.name, field.value


@ufunc.resolver(LoadBuilder, Bind, name='page_item')
def page_item(env, field):
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.name)
    return prop.name, prop


@ufunc.resolver(LoadBuilder, Positive, name='page_item')
def page_item(env, field):
    return env.call('page_item', Bind(field.name))


@ufunc.resolver(LoadBuilder, Negative, name='page_item')
def page_item(env, field):
    resolved = env.call('page_item', Bind(field.name))
    return f'-{resolved[0]}', resolved[1]


@ufunc.resolver(LoadBuilder, Expr, name='and')
def and_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return args


@ufunc.resolver(LoadBuilder, Expr, name='or')
def or_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return args


@ufunc.resolver(BaseQueryBuilder, DataType, list)
def validate_dtype_for_select(env, dtype: DataType, selected_props: List[Property]):
    raise NotImplemented(f"validate_dtype_for_select with {dtype.name} is not implemented")


@ufunc.resolver(BaseQueryBuilder, Text, list)
def validate_dtype_for_select(env, dtype: Text, selected_props: List[Property]):
    for prop in selected_props:
        if dtype.prop.name == prop.name or prop.parent == dtype.prop:
            raise CannotSelectTextAndSpecifiedLang(dtype)


@ufunc.resolver(BaseQueryBuilder, String, list)
def validate_dtype_for_select(env, dtype: String, selected_props: List[Property]):
    if dtype.prop.parent and isinstance(dtype.prop.parent.dtype, Text):
        parent = dtype.prop.parent
        for prop in selected_props:
            if parent.name == prop.name or prop == dtype.prop:
                raise CannotSelectTextAndSpecifiedLang(parent.dtype)


def filter_page_values(page: Page):
    new_page = Page()
    for by, page_by in page.by.items():
        if page_by.value is not None:
            new_page.by[by] = page_by
    return new_page


def _get_pagination_compare_query(
    model_page: Page,
) -> Union[Expr, None]:
    filtered = filter_page_values(model_page)
    item_count = len(filtered.by.keys())
    where_list = []
    for i in range(item_count):
        where_list.append([])
    for i, (by, page_by) in enumerate(filtered.by.items()):
        if page_by.value is not None:
            for n in range(item_count):
                if n >= i:
                    if n == i:
                        if by.startswith('-'):
                            where_list[n].append(('lt', page_by))
                        else:
                            where_list[n].append(('gt', page_by))
                    else:
                        where_list[n].append(('eq', page_by))

    where_compare = {}
    for where in where_list:
        compare = {}
        for item in where:
            if compare:
                compare = {
                    'name': 'and',
                    'args': [
                        compare,
                        {
                            'name': item[0],
                            'args': [{
                                'name': 'bind',
                                'args': [item[1].prop.name]
                            }, item[1].value]
                        }
                    ]
                }
            else:
                compare = {
                    'name': item[0],
                    'args': [{
                        'name': 'bind',
                        'args': [item[1].prop.name]
                    }, item[1].value]
                }
        if where_compare:
            where_compare = {
                'name': 'or',
                'args': [
                    where_compare,
                    compare
                ]
            }
        else:
            where_compare = compare

    if where_compare:
        where_compare = asttoexpr(where_compare)

    query = where_compare or None
    return query


def add_page_expr(expr: Expr, page: Page):
    return merge_formulas(expr, asttoexpr({
        'name': 'paginate',
        'args': [
            page
        ]
    }))



from uuid import UUID

import ujson as json
from decimal import Decimal

from spinta.formats.components import Format


class Json(Format):
    content_type = 'application/json'
    accept_types = {
        'application/json',
    }
    params = {}
    container_name = '_data'

    def __call__(self, data):
        yield f'{{"{self.container_name}":['
        for i, row in enumerate(data):
            sep = ',' if i > 0 else ''
            yield sep + json.dumps(self.data(row), default=encoder, ensure_ascii=False)
        yield ']}'

    def data(self, data: dict) -> dict:
        return data


def encoder(o):
    if isinstance(o, UUID):
        return str(o)
    if isinstance(o, Decimal):
        return float(o)




from spinta import commands
from spinta.components import Context, Model, Property
from spinta.exceptions import ModelReferenceNotFound, MultipleBackRefReferencesFound, NoBackRefReferencesFound, \
    NoReferencesFound, OneToManyBackRefNotSupported
from spinta.types.datatype import BackRef, Ref, Array, ArrayBackRef, Object, Denorm, DataType
from spinta.types.helpers import set_dtype_backend


@commands.find_backref_ref.register(Model, str, object)
def find_backref_ref(model: Model, backref_model: str, given_ref: object):
    for prop in model.properties.values():
        result = commands.find_backref_ref(prop, backref_model, given_ref)
        if result is not None:
            yield from result


@commands.find_backref_ref.register(Property, str, object)
def find_backref_ref(prop: Property, backref_model: str, given_ref: object):
    result = commands.find_backref_ref(prop.dtype, backref_model, given_ref)
    if result is not None:
        yield from result


@commands.find_backref_ref.register(DataType, str, object)
def find_backref_ref(dtype: DataType, backref_model: str, given_ref: str):
    pass


@commands.find_backref_ref.register(Ref, str, object)
def find_backref_ref(dtype: Ref, backref_model: str, given_ref: object):
    target_model = dtype.model.name if isinstance(dtype.model, Model) else dtype.model
    if target_model == backref_model:
        if given_ref and given_ref == dtype.prop.name or given_ref is None:
            yield dtype.prop


@commands.find_backref_ref.register(Array, str, object)
def find_backref_ref(dtype: Array, backref_model: str, given_ref: object):
    if dtype.items:
        result = commands.find_backref_ref(dtype.items.dtype, backref_model, given_ref)
        if result is not None:
            yield from result


@commands.find_backref_ref.register(Object, str, object)
def find_backref_ref(dtype: Object, backref_model: str, given_ref: object):
    for obj_prop in dtype.properties.values():
        result = commands.find_backref_ref(obj_prop.dtype, backref_model, given_ref)
        if result is not None:
            yield from result


@commands.find_backref_ref.register(Denorm, str, object)
def find_backref_ref(dtype: Denorm, backref_model: str, given_ref: object):
    result = commands.find_backref_ref(dtype.rel_prop.dtype, backref_model, given_ref)
    if result is not None:
        yield from result


def _link_backref(context: Context, dtype: BackRef):
    set_dtype_backend(dtype)
    backref_model = dtype.prop.model.name if isinstance(dtype.prop.model, Model) else dtype.prop.model
    backref_target_model = dtype.model.name if isinstance(dtype.model, Model) else dtype.model
    if backref_target_model == backref_model:
        # Self reference.
        dtype.model = dtype.prop.model
    else:
        if backref_target_model not in dtype.prop.model.manifest.models:
            raise ModelReferenceNotFound(dtype, ref=backref_target_model)
        dtype.model = dtype.prop.model.manifest.models[backref_target_model]
    given_refprop = dtype.refprop
    if dtype.refprop:
        dtype.explicit = True

    count = 0
    result = list(commands.find_backref_ref(dtype.model, backref_model, given_refprop))
    if not result:
        if given_refprop is not None:
            raise NoReferencesFound(dtype, prop_name=given_refprop, model=dtype.model)
        raise NoBackRefReferencesFound(dtype, model=dtype.model.name)
    for prop in result:
        if count < 1:
            if given_refprop and prop.name == given_refprop or given_refprop is None:
                dtype.refprop = prop
                count += 1
        else:
            raise MultipleBackRefReferencesFound(dtype, model=dtype.model.name)


@commands.link.register(Context, BackRef)
def link(context: Context, dtype: BackRef) -> None:
    _link_backref(context, dtype)

    if dtype.refprop.list is not None:
        raise OneToManyBackRefNotSupported(dtype)

    # relationship needs to add unique
    dtype.refprop.dtype.unique = True


@commands.link.register(Context, ArrayBackRef)
def link(context: Context, dtype: ArrayBackRef):
    _link_backref(context, dtype)



from typing import List
from typing import Optional

from spinta.components import Component
from spinta.core.enums import Access
from spinta.dimensions.comments.components import Comment
from spinta.dimensions.comments.components import CommentGiven
from spinta.manifests.tabular.components import CommentRow
from spinta.utils.enums import enum_by_name


def load_comments(
    parent: Component,
    data: Optional[List[CommentRow]],
) -> List[Comment]:
    comments = []
    if data:
        for comment_data in data:
            params = comment_data.copy()
            access_given = params.pop('access')
            access = enum_by_name(parent, 'access', Access, access_given)
            access = access or Access.private
            comments.append(Comment(
                access=access,
                given=CommentGiven(
                    access=access_given,
                ),
                **params,
            ))
    return comments



from typing import Union, List, Tuple

import itertools

from spinta import commands
from spinta.utils.schema import NA
from spinta.utils.data import take
from spinta.components import Context, Action, Property
from spinta.types.datatype import DataType, Object, Array
from spinta.components import Context, Action, Model, DataSubItem
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL


@commands.before_write.register(Context, Array, PostgreSQL)
def before_write(
    context: Context,
    dtype: Array,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
):
    if data.saved and data.patch is not NA:
        prop = dtype.prop
        table = backend.get_table(prop, TableType.LIST)
        transaction = context.get('transaction')
        connection = transaction.connection
        connection.execute(
            table.delete().
            where(table.c._rid == data.root.saved['_id'])
        )

    if dtype.prop.list:
        return {}
    else:
        return take({dtype.prop.place: data.patch})


@commands.after_write.register(Context, Array, PostgreSQL)
def after_write(
    context: Context,
    dtype: Array,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
):
    if data.patch:
        prop = dtype.prop
        table = backend.get_table(prop, TableType.LIST)
        transaction = context.get('transaction')
        connection = transaction.connection
        rid = take('_id', data.root.patch, data.root.saved)
        rows = [
            {
                _get_list_column_name(prop.place, k): v
                for k, v in commands.before_write(
                    context,
                    dtype.items.dtype,
                    backend,
                    data=d,
                ).items()
            }
            for d in data.iter(patch=True)
        ]
        qry = table.insert().values({
            '_txn': transaction.id,
            '_rid': rid,
        })
        connection.execute(qry, rows)

        for d in data.iter(patch=True):
            commands.after_write(context, dtype.items.dtype, backend, data=d)


def _update_lists_table(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    action: Action,
    pk: str,
    patch: dict,
) -> None:
    transaction = context.get('transaction')
    connection = transaction.connection
    rows = _get_lists_data(model, patch)
    sort_key = lambda x: x[0].place  # noqa
    rows = sorted(rows, key=sort_key)
    for place, rows in itertools.groupby(rows, key=sort_key):
        prop = model.flatprops[place]
        table = backend.get_table(prop, TableType.LIST)
        if action != Action.INSERT:
            connection.execute(table.delete().where(table.c._rid == pk))
        rows = [
            {
                '_txn': transaction.id,
                '_rid': pk,
                **{
                    _get_list_column_name(place, k): v
                    for k, v in row.items()
                }
            }
            for prop, row in rows
        ]
        connection.execute(table.insert(), rows)


def _get_list_column_name(place, name):
    if place == name:
        return place.split('.')[-1]
    else:
        return name[len(place) + 1:]



def _get_lists_data(
    dtype: Union[Model, DataType],
    value: object,
) -> List[dict]:
    data, lists = _separate_lists_and_data(dtype, value)
    if isinstance(dtype, DataType) and data is not NA:
        yield dtype.prop, data
    for prop, vals in lists:
        for v in vals:
            yield from _get_lists_data(prop.dtype, v)


def _separate_lists_and_data(
    dtype: Union[Model, DataType],
    value: object,
) -> Tuple[dict, List[Tuple[Property, list]]]:
    if isinstance(dtype, (Model, Object)):
        data = {}
        lists = []
        for k, v in (value or {}).items():
            prop = dtype.properties[k]
            v, more = _separate_lists_and_data(prop.dtype, v)
            if v is not NA:
                data.update(v)
            if more:
                lists += more
        return data or NA, lists
    elif isinstance(dtype, Array):
        if value:
            return NA, [(dtype.items, value)]
        else:
            return NA, []
    else:
        return {dtype.prop.place: value}, []



from spinta.backends.nobackend.components import NoBackend
from spinta.exceptions import BackendNotGiven
from spinta import commands
from spinta.core.ufuncs import Expr
from spinta.components import Context, UrlParams
from spinta.components import Model


@commands.getall.register(Context, Model, NoBackend)
def getall(
    context: Context,
    model: Model,
    backend: NoBackend,
    *,
    query: Expr = None,
    **kwargs
):
    raise BackendNotGiven(model)


@commands.getone.register(Context, Model, NoBackend)
def getone(
    context: Context,
    model: Model,
    db: NoBackend,
    *,
    id_: str,
):
    raise BackendNotGiven(model)


@commands.changes.register(Context, Model, NoBackend)
def changes(
    context: Context,
    model: Model,
    db: NoBackend,
    *,
    id_: str = None,
    limit: int = 100,
    offset: int = -10,
):
    raise BackendNotGiven(model)



from typing import Iterator
from typing import overload

from spinta.typing import ObjectData
from spinta.typing import FileObjectData
from spinta import commands
from spinta.components import Context, UrlParams
from spinta.components import Model
from spinta.components import Property
from spinta.core.ufuncs import Expr
from spinta.types.datatype import File, Object
from spinta.exceptions import ItemDoesNotExist
from spinta.backends.mongo.components import Mongo
from spinta.backends.mongo.commands.query import MongoQueryBuilder
from spinta.ufuncs.basequerybuilder.components import get_page_values


@overload
@commands.getone.register(Context, Model, Mongo)
def getone(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    id_: str,
) -> ObjectData:
    table = backend.db[model.model_type()]
    keys = {k: 1 for k in model.flatprops}
    keys['__id'] = 1
    keys['_id'] = 0
    query = {'__id': id_}
    data = table.find_one(query, keys)
    if data is None:
        raise ItemDoesNotExist(model, id=id_)
    data['_id'] = data['__id']
    data['_type'] = model.model_type()
    return commands.cast_backend_to_python(context, model, backend, data)


@overload
@commands.getone.register(Context, Property, Object, Mongo)
def getone(
    context: Context,
    prop: Property,
    dtype: Object,
    backend: Mongo,
    *,
    id_: str,
) -> ObjectData:
    table = backend.db[prop.model.model_type()]
    data = table.find_one({'__id': id_}, {
        '__id': 1,
        '_revision': 1,
        prop.name: 1,
    })
    if data is None:
        raise ItemDoesNotExist(prop, id=id_)
    result = {
        '_id': data['__id'],
        '_revision': data['_revision'],
        '_type': prop.model_type(),
        prop.name: (data.get(prop.name) or {}),
    }
    return commands.cast_backend_to_python(context, prop, backend, result)


@overload
@commands.getone.register(Context, Property, File, Mongo)
def getone(
    context: Context,
    prop: Property,
    dtype: File,
    backend: Mongo,
    *,
    id_: str,
) -> FileObjectData:
    table = backend.db[prop.model.model_type()]
    data = table.find_one({'__id': id_}, {
        '__id': 1,
        '_revision': 1,
        prop.name: 1,
    })
    if data is None:
        raise ItemDoesNotExist(prop, id=id_)
    # merge file property data with defaults
    file_data = {
        **{
            '_content_type': None,
            '_id': None,
            '_size': None,
        },
        **data.get(prop.name, {}),
    }
    result = {
        '_id': data['__id'],
        '_revision': data['_revision'],
        '_type': prop.model_type(),
        prop.name: file_data,
    }
    return commands.cast_backend_to_python(context, prop, backend, result)


@commands.getall.register(Context, Model, Mongo)
def getall(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    query: Expr = None,
    **kwargs
) -> Iterator[ObjectData]:
    builder = MongoQueryBuilder(context)
    builder.update(model=model)
    table = backend.db[model.model_type()]
    env = builder.init(backend, table)
    expr = env.resolve(query)
    where = env.execute(expr)
    cursor = env.build(where)
    for row in cursor:
        if '__id' in row:
            row['_id'] = row.pop('__id')
        row['_type'] = model.model_type()
        if model.page.is_enabled:
            row['_page'] = get_page_values(env, row)
        yield commands.cast_backend_to_python(context, model, backend, row)



from spinta import commands
from spinta.components import Property
from spinta.datasets.backends.sql.components import Sql
from spinta.exceptions import NoExternalName, PropertyNotFound
from spinta.types.datatype import DataType
from spinta.types.text.components import Text
import sqlalchemy as sa


@commands.get_column.register(Sql, Property)
def get_column(backend: Sql, prop: Property, **kwargs):
    return commands.get_column(backend, prop.dtype, **kwargs)


@commands.get_column.register(Sql, DataType)
def get_column(backend: Sql, dtype: DataType, table: sa.Table, **kwargs):
    prop = dtype.prop
    if prop.external is None or not prop.external.name:
        raise NoExternalName(prop)
    if prop.external.name not in table.c:
        raise PropertyNotFound(
            prop.model,
            property=prop.name,
            external=prop.external.name,
        )
    return table.c[prop.external.name]


@commands.get_column.register(Sql, Text)
def get_column(backend: Sql, dtype: Text, table: sa.Table, langs: list = [], default_langs: list = [], push: bool = False, **kwargs):
    prop = dtype.prop

    existing_langs = list(dtype.langs.keys())
    lang_prop = None
    if langs:
        for lang in langs:
            if lang in dtype.langs:
                lang_prop = dtype.langs[lang]
                break
    if not lang_prop and default_langs:
        for lang in default_langs:
            if lang in dtype.langs:
                lang_prop = dtype.langs[lang]
                break
    if not lang_prop:
        lang_prop = dtype.langs[existing_langs[0]]

    column_name = lang_prop.external.name
    if not column_name:
        raise NoExternalName(lang_prop)
    if column_name not in table.c:
        raise PropertyNotFound(
            prop.model,
            property=lang_prop.name,
            external=lang_prop.external.name,
        )
    column = table.c[lang_prop.external.name]
    return column



from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.dataframe.ufuncs.components import TabularResource
from spinta.exceptions import UnknownBind


@ufunc.resolver(TabularResource)
def tabular(env: TabularResource, **kwargs):
    for key, value in kwargs.items():
        if key == "sep":
            env.seperator = value
        else:
            raise UnknownBind(name=key)



from spinta import commands
from spinta.components import Context
from spinta.manifests.backend.components import BackendManifest


@commands.freeze.register(Context, BackendManifest)
def freeze(context: Context, current: BackendManifest):
    # Backend manifest delegates freezing to its sync manifests.
    for source in current.sync:
        commands.freeze(context, source)



import base64
import datetime
import pathlib
import decimal


def fix_data_for_json(data):
    if isinstance(data, dict):
        return {k: fix_data_for_json(v) for k, v in data.items()}
    if isinstance(data, list):
        return [fix_data_for_json(v) for v in data]
    if isinstance(data, tuple):
        return (fix_data_for_json(v) for v in data)
    if isinstance(data, (datetime.datetime, datetime.date, datetime.time)):
        return data.isoformat()
    if isinstance(data, bytes):
        return base64.b64encode(data).decode('ascii')
    if isinstance(data, pathlib.Path):
        return str(data)
    if isinstance(data, decimal.Decimal):
        return float(data)
    if isinstance(data, (int, float, str, type(None))):
        return data
    raise TypeError(f"{type(data)} probably won't serialize to JSON.")



from spinta import commands, exceptions
import spinta.backends.postgresql.helpers.extractors as extractor
from spinta.components import DataItem, Action
from sqlalchemy.exc import IntegrityError
import psycopg2.errors as psy_errors


@commands.create_exception.register(DataItem, IntegrityError)
def create_exception(item: DataItem, error: IntegrityError):
    return create_exception(item, error.orig)


@commands.create_exception.register(DataItem, psy_errors.ForeignKeyViolation)
def create_exception(item: DataItem, error: psy_errors.ForeignKeyViolation):
    error_message = error.diag.message_detail
    if item.action is Action.DELETE:
        error_model = extractor.extract_error_model(error_message)
        return exceptions.ReferringObjectFound(
            item.model.properties.get("_id"),
            model=error_model,
            id=item.saved.get("_id"))
    else:
        error_property_names = extractor.extract_error_property_names(error_message)
        error_ref_id = extractor.extract_error_ref_id(error_message)
        context = item.model
        if len(error_property_names) == 1:
            context = item.model.properties.get(error_property_names[0])
        return exceptions.ReferencedObjectNotFound(
            context,
            id=error_ref_id)


@commands.create_exception.register(DataItem, psy_errors.UniqueViolation)
def create_exception(item: DataItem, error: psy_errors.UniqueViolation):
    error_message = error.diag.message_detail
    error_property_names = extractor.extract_error_property_names(error_message)
    if len(error_property_names) == 1:
        error_property = item.model.properties.get(error_property_names[0])
        return exceptions.UniqueConstraint(error_property)
    else:
        return exceptions.CompositeUniqueConstraint(
            item.model,
            properties=",".join(error_property_names)
        )


@commands.create_exception.register(DataItem, Exception)
def create_exception(item: DataItem, error: Exception):
    raise error



# Maximum length for PostgreSQL identifiers (e.g. table names, column names,
# function names).
# https://github.com/postgres/postgres/blob/master/src/include/pg_config_manual.h
NAMEDATALEN = 63

UNSUPPORTED_TYPES = [
    'generic',
    'rql',
    'inherit'
]



from typing import Any
from typing import List

import os
import sys
import traceback

import pytest
from typer.testing import CliRunner

from spinta.cli import main
from spinta.core.config import RawConfig
from spinta.testing.context import create_test_context


def _prepare_args(args: List[Any]) -> List[str]:
    """Prepare args

    This does following conversions:

        [['a']] -> ['a']
        [1]     -> ['1']
        [None]  -> ['']

    """
    if not args:
        return args

    result = []
    for arg in args:
        if not isinstance(arg, list):
            arg = [arg]
        result += ['' if a is None else str(a) for a in arg]
    return result


class SpintaCliRunner(CliRunner):

    def invoke(
        self,
        rc: RawConfig,
        args: List[Any] = None,
        fail: bool = True,
        **kwargs,
    ):
        assert isinstance(rc, RawConfig)
        if 'obj' not in kwargs:
            context = create_test_context(rc, name='pytest/cli')
            kwargs['obj'] = context

        args = _prepare_args(args)
        result = super().invoke(main.app, args, **kwargs)
        if result.exc_info is not None:
            t, e, tb = result.exc_info
            if not isinstance(e, SystemExit):
                exc = ''.join(traceback.format_exception(t, e, tb))
                exc = exc.replace(os.getcwd() + '/', '')
                print(exc, file=sys.stderr)
        if result.exit_code != 0:
            print(result.stdout)
            print(result.stderr, file=sys.stderr)
            cmd = ' '.join(['spinta'] + (args or []))
            if fail:
                pytest.fail(f"Command `{cmd}` failed, exit code {result.exit_code}.")
        else:
            if result.stderr_bytes is not None and result.stderr:
                print(result.stderr, file=sys.stderr)
        return result



import cgi
import pathlib
import typing

from starlette.exceptions import HTTPException
from starlette.requests import Request

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.accesslog import log_async_response

from spinta.renderer import render
from spinta.utils.aiotools import aiter
from spinta.components import Context, Action, UrlParams, DataItem
from spinta.commands.write import prepare_patch, simple_response, validate_data
from spinta.types.datatype import File
from spinta.backends.fs.components import FileSystem

if typing.TYPE_CHECKING:
    from spinta.backends.postgresql.components import WriteTransaction


@commands.push.register()
async def push(
    context: Context,
    request: Request,
    dtype: File,
    backend: FileSystem,
    *,
    action: Action,
    params: UrlParams,
):
    prop = dtype.prop

    commands.authorize(context, action, prop)

    transaction: WriteTransaction = context.get('transaction')
    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
        rev=request.headers.get('revision'),
        txn=transaction.id,
    )

    data = DataItem(
        prop.model,
        prop,
        propref=False,
        backend=backend,
        action=action
    )
    data.given = {
        prop.name: {
            '_content_type': request.headers.get('content-type'),
            '_id': None,
        }
    }

    if 'revision' in request.headers:
        data.given['_revision'] = request.headers.get('revision')
    if 'content-disposition' in request.headers:
        data.given[prop.name]['_id'] = cgi.parse_header(request.headers['content-disposition'])[1]['filename']
    require_content_length = (
        Action.INSERT,
        Action.UPSERT,
        Action.UPDATE,
        Action.PATCH,
    )
    if action in require_content_length and 'content-length' not in request.headers:
        raise HTTPException(status_code=411)
    if not data.given[prop.name]['_id']:
        # XXX: Probably here should be a new UUID.
        data.given[prop.name]['_id'] = params.pk

    commands.simple_data_check(context, data, data.prop, data.model.backend)

    data.saved = commands.getone(context, prop, dtype, prop.model.backend, id_=params.pk)

    filepath = backend.path / data.given[prop.name]['_id']

    dstream = aiter([data])
    dstream = validate_data(context, dstream)
    dstream = prepare_patch(context, dstream)
    if action in (Action.UPDATE, Action.PATCH):
        dstream = create_file(filepath, dstream, request.stream())
        dstream = commands.update(context, prop, dtype, prop.model.backend, dstream=dstream)
    elif action == Action.DELETE:
        dstream = commands.delete(context, prop, dtype, prop.model.backend, dstream=dstream)
    else:
        raise Exception(f"Unknown action {action!r}.")
    dstream = commands.create_changelog_entry(
        context, prop.model, prop.model.backend, dstream=dstream,
    )

    dstream = log_async_response(context, dstream)

    status_code, response = await simple_response(context, params.fmt, dstream)
    return render(context, request, prop, params, response, action=action, status_code=status_code)


async def create_file(
    filepath: pathlib.PosixPath,
    dstream: typing.AsyncIterator[DataItem],
    fstream: typing.AsyncIterator[bytes],
) -> typing.AsyncGenerator[DataItem, None]:
    # we know that dstream contains one DataItem element
    async for d in dstream:
        with open(filepath, 'wb') as f:
            async for chunk in fstream:
                f.write(chunk)
        yield d



from spinta import commands
from spinta.backends.nobackend.components import NoBackend
from spinta.components import Context


@commands.bootstrap.register(Context, NoBackend)
def bootstrap(context: Context, backend: NoBackend):
    pass



from spinta import commands
from spinta.components import Context
from spinta.datasets.components import ExternalBackend
from spinta.datasets.components import Entity


@commands.wipe.register(Context, Entity, ExternalBackend)
def wipe(context: Context, entity: Entity, backend: ExternalBackend):
    raise NotImplementedError



import uuid
from copy import deepcopy
from typing import overload, Optional, Iterator, List, Tuple
from pathlib import Path

from starlette.requests import Request
from starlette.responses import Response
from starlette.responses import FileResponse

from spinta import commands
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.backends.components import Backend
from spinta.backends.nobackend.components import NoBackend
from spinta.compat import urlparams_to_expr
from spinta.components import Context, Node, Action, UrlParams, Page, PageBy
from spinta.components import Model
from spinta.components import Property
from spinta.core.ufuncs import Expr
from spinta.datasets.components import ExternalBackend
from spinta.renderer import render
from spinta.types.datatype import Integer
from spinta.types.datatype import Object
from spinta.types.datatype import File
from spinta.accesslog import AccessLog
from spinta.accesslog import log_response
from spinta.exceptions import UnavailableSubresource, InfiniteLoopWithPagination, DuplicateRowWhilePaginating, BackendNotGiven
from spinta.exceptions import ItemDoesNotExist
from spinta.types.datatype import DataType
from spinta.typing import ObjectData
from spinta.ufuncs.basequerybuilder.components import get_allowed_page_property_types, QueryParams, \
    update_query_with_url_params
from spinta.ufuncs.basequerybuilder.ufuncs import add_page_expr
from spinta.utils.data import take
from spinta.utils.encoding import encode_page_values


@commands.getall.register(Context, Model, Request)
async def getall(
    context: Context,
    model: Model,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)
    backend = model.backend
    if isinstance(backend, NoBackend):
        raise BackendNotGiven(model)

    copy_page = prepare_page_for_get_all(context, model, params)
    is_page_enabled = not params.count and backend.paginated and copy_page and copy_page and copy_page.is_enabled

    if isinstance(backend, ExternalBackend):
        # XXX: `add_count` is a hack, because, external backends do not
        #      support it yet.
        expr = urlparams_to_expr(params, add_count=False)
    else:
        expr = urlparams_to_expr(params)
    query_params = QueryParams()
    update_query_with_url_params(query_params, params)
    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
    )

    if params.head:
        rows = []
    else:
        if is_page_enabled:
            rows = get_page(context, model, backend, copy_page, expr, params.limit, default_expand=False, params=query_params)
        else:
            if backend.support_expand:
                rows = commands.getall(context, model, backend, params=query_params, query=expr, default_expand=False)
            else:
                rows = commands.getall(context, model, backend, params=query_params, query=expr)

    if params.count:
        # XXX: Quick and dirty hack. Functions should be handled properly.
        prop = Property()
        prop.name = 'count()'
        prop.place = 'count()'
        prop.title = ''
        prop.description = ''
        prop.model = model
        prop.dtype = Integer()
        prop.dtype.type = 'integer'
        prop.dtype.type_args = []
        prop.dtype.name = 'integer'
        prop.dtype.prop = prop
        props = {
            '_type': model.properties['_type'],
            'count()': prop,
        }
        rows = (
            {
                prop.name: commands.prepare_dtype_for_response(
                    context,
                    params.fmt,
                    props[key].dtype,
                    val,
                    data=row,
                    action=action,
                )
                for key, val in row.items() if key in props
            }
            for row in rows
        )
    else:
        select_tree = get_select_tree(context, action, params.select)
        if action == Action.SEARCH:
            reserved = ['_type', '_id', '_revision', '_base']
        else:
            reserved = ['_type', '_id', '_revision']
        if model.page.is_enabled:
            reserved.append('_page')
        prop_names = get_select_prop_names(
            context,
            model,
            model.properties,
            action,
            select_tree,
            reserved=reserved,
            include_denorm_props=False,
        )
        rows = (
            commands.prepare_data_for_response(
                context,
                model,
                params.fmt,
                row,
                action=action,
                select=select_tree,
                prop_names=prop_names,
            )
            for row in rows
        )

    rows = log_response(context, rows)

    return render(context, request, model, params, rows, action=action)


# XXX: params.sort handle should be moved to BaseQueryBuilder, AST should be handled by Env, this only supports basic sort arguments, that ar Positive, Negative or Bind.
# Issue: in order to create correct WHERE, we need to have finalized Page, currently we combine sort properties with page here.
def prepare_page_for_get_all(context: Context, model: Model, params: UrlParams):
    if model.page:
        copied = deepcopy(model.page)
        copied.clear()

        if params.page:
            copied.is_enabled = params.page.is_enabled
        if copied.is_enabled:
            config = context.get('config')
            page_size = config.push_page_size
            size = params.page and params.page.size or copied.size or page_size or 1000
            copied.size = size

            if params.sort:
                new_order = {}
                for sort_ast in params.sort:
                    negative = False
                    if sort_ast['name'] == 'negative':
                        negative = True
                        sort_ast = sort_ast['args'][0]
                    elif sort_ast['name'] == 'positive':
                        sort_ast = sort_ast['args'][0]
                    for sort in sort_ast['args']:
                        if isinstance(sort, str):
                            if sort in model.properties.keys():
                                by = sort if not negative else f'-{sort}'
                                prop = model.properties[sort]
                                if isinstance(prop.dtype, get_allowed_page_property_types()):
                                    new_order[by] = PageBy(prop)
                                else:
                                    copied.is_enabled = False
                                    break
                        else:
                            copied.is_enabled = False
                            break

                for key, page_by in copied.by.items():
                    reversed_key = key[1:] if key.startswith("-") else f'-{key}'
                    if key not in new_order and reversed_key not in new_order:
                        new_order[key] = page_by

                copied.by = new_order

            if params.page and params.page.values:
                copied.update_values_from_list(params.page.values)

        return copied


def get_page(
    context: Context,
    model: Model,
    backend: Backend,
    model_page: Page,
    expr: Expr,
    limit: Optional[int] = None,
    default_expand: bool = True,
    params: QueryParams = None,
) -> Iterator[ObjectData]:
    config = context.get('config')
    page_size = config.push_page_size
    model_page.size = model_page.size or page_size or 1000

    true_count = 0

    last_value = None
    while True:
        finished = True
        query = add_page_expr(expr, model_page)
        if backend.support_expand:
            rows = commands.getall(context, model, backend, params=params, query=query, default_expand=default_expand)
        else:
            rows = commands.getall(context, model, backend, params=params, query=query)
        first_value = None
        previous_value = None
        for row in rows:
            if previous_value is not None:
                if previous_value == row:
                    raise DuplicateRowWhilePaginating(key=encode_page_values(row['_page']))
            previous_value = row
            if finished:
                finished = False

            if limit and true_count >= limit:
                finished = True
                break

            if first_value is None:
                first_value = row
                if first_value == last_value:
                    raise InfiniteLoopWithPagination()
                else:
                    last_value = first_value

            true_count += 1
            if '_page' in row:
                model_page.update_values_from_list(row['_page'])
            yield row

        if finished:
            break


def _update_expr_args(
    expr: Expr,
    name: str,
    args: List,
    override: bool = False
) -> Tuple[bool, Expr]:
    updated = False
    expr.args = list(expr.args)
    if expr.name == name:
        if override:
            expr.args = args
            updated = True
        else:
            if expr.args:
                expr.args.extend(args)
                updated = True
    else:
        for i, arg in enumerate(expr.args):
            if isinstance(arg, Expr):
                updated, expr.args[i] = _update_expr_args(arg, name, args, override)
                if updated:
                    break
    expr.args = tuple(expr.args)
    return updated, expr


@overload
@commands.getone.register(Context, Request, Node)
async def getone(
    context: Context,
    request: Request,
    node: Node,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    if params.prop and params.propref:
        resp = await commands.getone(
            context,
            request,
            params.prop,
            params.prop.dtype,
            params.model.backend,
            action=action,
            params=params,
        )
    elif params.prop:
        resp = await commands.getone(
            context,
            request,
            params.prop,
            params.prop.dtype,
            params.prop.dtype.backend or params.model.backend,
            action=action,
            params=params,
        )
    else:
        resp = await commands.getone(
            context,
            request,
            params.model,
            params.model.backend,
            action=action,
            params=params,
        )

    accesslog: AccessLog = context.get('accesslog')
    accesslog.response(objects=1)

    return resp


@overload
@commands.getone.register(Context, Request, Property, DataType, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    raise UnavailableSubresource(prop=prop.name, prop_type=prop.dtype.name)


@overload
@commands.getone.register(Context, Request, Model, Backend)
async def getone(
    context: Context,
    request: Request,
    model: Model,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
        id_=params.pk,
    )

    data = commands.getone(context, model, backend, id_=params.pk)
    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        select_tree,
    )
    data = commands.prepare_data_for_response(
        context,
        model,
        params.fmt,
        data,
        action=action,
        select=select_tree,
        prop_names=prop_names,
    )
    return render(context, request, model, params, data, action=action)


@overload
@commands.getone.register(Context, Request, Property, Object, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: Object,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    data = commands.getone(
        context,
        prop,
        dtype,
        backend,
        id_=params.pk,
    )

    data = commands.prepare_data_for_response(
        context,
        prop.dtype,
        params.fmt,
        data,
        action=action,
    )
    return render(context, request, prop, params, data, action=action)


@overload
@commands.getone.register(Context, Request, Property, File, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: File,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    # Get metadata from model backend
    data = commands.getone(
        context,
        prop,
        dtype,
        prop.model.backend,
        id_=params.pk,
    )

    # Return file metadata from model backend
    if params.propref:
        data = commands.prepare_data_for_response(
            context,
            prop.dtype,
            params.fmt,
            data,
            action=Action.GETONE,
        )
        return render(context, request, prop, params, data, action=action)

    # Return file content from property backend
    else:
        value = take(prop.place, data)

        file = commands.getfile(
            context,
            prop,
            dtype,
            dtype.backend,
            data=value,
        )

        if file is None:
            raise ItemDoesNotExist(dtype, id=params.pk)

        filename = value['_id']

        if isinstance(file, bytes):
            ResponseClass = Response
        elif isinstance(file, Path):
            ResponseClass = FileResponse
        else:
            raise ValueError(f"Unknown file type {type(file)}.")

        return ResponseClass(
            file,
            media_type=value.get('_content_type'),
            headers={
                'Revision': data['_revision'],
                'Content-Disposition': (
                    f'attachment; filename="{filename}"'
                    if filename else
                    'attachment'
                )
            },
        )


@commands.changes.register(Context, Model, Request)
async def changes(
    context: Context,
    model: Model,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)

    if params.head:
        rows = []
    else:
        accesslog = context.get('accesslog')
        accesslog.request(
            # XXX: Read operations does not have a transaction, but it
            #      is needed for loging.
            txn=str(uuid.uuid4()),
            model=model.model_type(),
            action=action.value,
        )

        rows = commands.changes(
            context,
            model,
            model.backend,
            id_=params.pk,
            limit=params.limit,
            offset=params.changes_offset,
        )

    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        select_tree,
        reserved=[
            '_cid',
            '_created',
            '_op',
            '_id',
            '_txn',
            '_revision',
        ],
    )
    rows = (
        commands.prepare_data_for_response(
            context,
            model,
            params.fmt,
            row,
            action=action,
            select=select_tree,
            prop_names=prop_names,
        )
        for row in rows
    )

    return render(context, request, model, params, rows, action=action)


@commands.summary.register(Context, Request, Model)
async def summary(
    context: Context,
    request: Request,
    model: Model,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)
    backend = model.backend

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
    )
    if params.head:
        rows = []
    else:
        expr = urlparams_to_expr(params)
        rows = commands.summary(context, model, backend, expr)
    rows = log_response(context, rows)
    return render(context, request, model, params, rows, action=action)



from spinta import commands
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.helpers import check_manifest_path
from spinta.manifests.dict.components import DictManifest


@commands.configure.register(Context, DictManifest)
def configure(context: Context, manifest: DictManifest):
    rc: RawConfig = context.get('rc')
    path = rc.get('manifests', manifest.name, 'path')
    if path:
        check_manifest_path(manifest, path)
    else:
        path = None
    manifest.path = path



from spinta.formats.components import Format


class Rdf(Format):
    content_type = 'application/rdf+xml'
    accept_types = {
        'application/rdf+xml',
    }
    params = {}
    prioritize_uri = True




from spinta.manifests.components import Manifest


class RdfManifest(Manifest):
    type: str = 'rdf'
    format: str

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return path.endswith('.rdf') or path.endswith('.ttl')



from typing import Iterator

import pathlib

from spinta import commands
from spinta.components import Context
from spinta.manifests.backend.components import BackendManifest
from spinta.manifests.backend.helpers import list_schemas
from spinta.manifests.backend.helpers import read_schema
from spinta.manifests.backend.helpers import get_last_version_eid
from spinta.manifests.backend.helpers import get_version_schema


@commands.manifest_list_schemas.register(Context, BackendManifest)
def manifest_list_schemas(
    context: Context,
    manifest: BackendManifest,
) -> Iterator[str]:
    yield from list_schemas(context, manifest)


@commands.manifest_read_current.register(Context, BackendManifest)
def manifest_read_current(
    context: Context,
    manifest: BackendManifest,
    *,
    eid: str,
) -> dict:
    return read_schema(context, manifest, eid)


@commands.manifest_read_freezed.register(Context, BackendManifest)
def manifest_read_freezed(
    context: Context,
    manifest: BackendManifest,
    *,
    eid: str,
) -> dict:
    version_eid = get_last_version_eid(context, manifest, eid)
    return get_version_schema(context, manifest, version_eid)


@commands.manifest_read_versions.register(Context, BackendManifest)
def manifest_read_versions(
    context: Context,
    manifest: BackendManifest,
    *,
    eid: pathlib.Path,
) -> Iterator[dict]:
    raise NotImplementedError



from datetime import datetime
from typing import Optional, Any

from spinta.components import Component


class KeyMap(Component):
    name: str = None

    def encode(self, name: str, value: Any, primary_key: Optional[str] = None) -> Optional[str]:
        raise NotImplementedError

    def decode(self, name: str, key: str) -> object:
        raise NotImplementedError

    def first_time_sync(self) -> bool:
        raise NotImplementedError

    def get_sync_data(self, name: str) -> object:
        raise NotImplementedError

    def update_sync_data(self, name: str, cid: Any, time: datetime):
        raise NotImplementedError

    def synchronize(self, name: str, value: Any, primary_key: str):
        raise NotImplementedError



from typing import List

import datetime
from typing import Union

from sqlalchemy.engine.row import RowProxy

from spinta import commands
from spinta.backends.components import SelectTree
from spinta.formats.components import Format
from spinta.components import Context, Model, Action
from spinta.types.datatype import Date, DateTime
from spinta.backends.postgresql.components import PostgreSQL


@commands.prepare_data_for_response.register(Context, Model, Format, RowProxy)
def prepare_data_for_response(
    context: Context,
    model: Model,
    fmt: Format,
    value: RowProxy,
    *,
    action: Action,
    select: SelectTree = None,
    prop_names: List[str] = None,
) -> dict:
    return commands.prepare_data_for_response(
        context,
        model,
        fmt,
        dict(value),
        action=action,
        select=select,
        prop_names=prop_names,
    )


@commands.prepare_for_write.register(Context, DateTime, PostgreSQL, datetime.datetime)
def prepare_for_write(
    context: Context,
    dtype: DateTime,
    backend: PostgreSQL,
    value: datetime.datetime,
) -> Union[datetime.datetime, str]:
    # convert datetime object to isoformat string if it belongs
    # to a nested property
    if dtype.prop.parent is dtype.prop.model:
        return value
    else:
        # XXX: Probably this should not be converted to string, because
        #      if I recall correctly, now nested objects are stored in separate
        #      tables.
        return value.isoformat()


@commands.prepare_for_write.register(Context, Date, PostgreSQL, datetime.date)
def prepare_for_write(
    context: Context,
    dtype: Date,
    backend: PostgreSQL,
    value: datetime.date,
) -> Union[datetime.date, str]:
    # convert date object to isoformat string if it belongs
    # to a nested property
    if dtype.prop.parent is dtype.prop.model:
        return value
    else:
        return value.isoformat()



from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.rdf.components import RdfManifest
from spinta.manifests.rdf.helpers import read_rdf_manifest


@commands.load.register(Context, RdfManifest)
def load(
    context: Context,
    manifest: RdfManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    rename_duplicates: bool = False,
    load_internal: bool = True,
):
    if load_internal:
        target = into or manifest
        if '_schema' not in target.models:
            store = context.get('store')
            commands.load(context, store.internal, into=target)

    if manifest.path is None:
        return

    schemas = read_rdf_manifest(manifest)

    if into:
        load_manifest_nodes(context, into, schemas, source=manifest)
    else:
        load_manifest_nodes(context, manifest, schemas)

    for source in manifest.sync:
        commands.load(
            context, source,
            into=into or manifest,
            freezed=freezed,
            rename_duplicates=rename_duplicates,
            load_internal=load_internal,
        )



from typing import Dict, Any, Tuple, List

from dask import dataframe as Dataframe

from spinta.auth import authorized
from spinta.components import Model, Property, Action
from spinta.core.ufuncs import Env, Expr, ufunc, Bind, Unresolved
from spinta.datasets.backends.dataframe.components import DaskBackend
from spinta.datasets.backends.sql.commands.query import GetAttr
from spinta.dimensions.param.components import ResolvedParams
from spinta.exceptions import UnknownMethod, PropertyNotFound, NotImplementedFeature, SourceCannotBeList
from spinta.types.datatype import DataType, PrimaryKey, Ref
from spinta.ufuncs.components import ForeignProperty
from spinta.utils.data import take
from spinta.utils.schema import NA


class Selected:
    # Item index in select list.
    item: str = None
    # Model property if a property is selected.
    prop: Property = None
    # A value or an Expr for further processing on selected value.
    prep: Any = NA

    def __init__(
        self,
        item: str = None,
        prop: Property = None,
        # `prop` can be Expr or any other value.
        prep: Any = NA,
    ):
        self.item = item
        self.prop = prop
        self.prep = prep

    def __repr__(self):
        return self.debug()

    def debug(self, indent: str = ''):
        prop = self.prop.place if self.prop else 'None'
        if isinstance(self.prep, Selected):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep=...)\n'
            ) + self.prep.debug(indent + '  ')
        elif isinstance(self.prep, (tuple, list)):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={type(self.prep).__name__}...)\n'
            ) + ''.join([
                p.debug(indent + '- ')
                if isinstance(p, Selected)
                else str(p)
                for p in self.prep
            ])
        else:
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={self.prep})\n'
            )


class DaskDataFrameQueryBuilder(Env):
    backend: DaskBackend
    model: Model
    dataframe: Dataframe
    # `resolved` is used to map which prop.place properties are already
    # resolved, usually it maps to Selected, but different DataType's can return
    # different results.
    resolved: Dict[str, Selected]
    selected: Dict[str, Selected] = None
    params: ResolvedParams

    def init(self, backend: DaskBackend, dataframe: Dataframe):
        return self(
            backend=backend,
            dataframe=dataframe,
            resolved={},
            selected=None,
            sort={
                "desc": [],
                "asc": []
            },
            limit=None,
            offset=None,
        )

    def build(self, where):
        if self.selected is None:
            self.call('select', Expr('select'))
        df = self.dataframe

        if self.limit is not None:
            df = df.loc[:self.limit - 1]
        #
        if self.offset is not None:
            df = df.loc[self.offset:]
        return df

    def execute(self, expr: Any):
        expr = self.call('_resolve_unresolved', expr)
        return super().execute(expr)

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownMethod(name=expr.name, expr=str(expr(*args, **kwargs)))


@ufunc.resolver(DaskDataFrameQueryBuilder, int)
def limit(env: DaskDataFrameQueryBuilder, n: int):
    env.limit = n


@ufunc.resolver(DaskDataFrameQueryBuilder, int)
def offset(env: DaskDataFrameQueryBuilder, n: int):
    env.offset = n


@ufunc.resolver(DaskDataFrameQueryBuilder, Bind, Bind, name='getattr')
def getattr_(env: DaskDataFrameQueryBuilder, obj: Bind, attr: Bind):
    return GetAttr(obj.name, attr)


@ufunc.resolver(DaskDataFrameQueryBuilder, Bind, GetAttr, name='getattr')
def getattr_(env: DaskDataFrameQueryBuilder, obj: Bind, attr: GetAttr):
    return GetAttr(obj.name, attr)


@ufunc.resolver(DaskDataFrameQueryBuilder, GetAttr)
def _resolve_getattr(
    env: DaskDataFrameQueryBuilder,
    attr: GetAttr,
) -> ForeignProperty:
    prop = env.model.properties[attr.obj]
    return env.call('_resolve_getattr', prop.dtype, attr.name)


@ufunc.resolver(DaskDataFrameQueryBuilder, Ref, GetAttr)
def _resolve_getattr(
    env: DaskDataFrameQueryBuilder,
    dtype: Ref,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = ForeignProperty(None, dtype, prop.dtype)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(DaskDataFrameQueryBuilder, Ref, Bind)
def _resolve_getattr(
    env: DaskDataFrameQueryBuilder,
    dtype: Ref,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, Ref, GetAttr)
def _resolve_getattr(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = fpr.push(prop)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, Ref, Bind)
def _resolve_getattr(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return fpr.push(prop)


@ufunc.resolver(DaskDataFrameQueryBuilder, object)
def _resolve_unresolved(env: DaskDataFrameQueryBuilder, value: Any) -> Any:
    if isinstance(value, Unresolved):
        raise ValueError(f"Unresolved value {value!r}.")
    else:
        return value


@ufunc.resolver(DaskDataFrameQueryBuilder, Bind)
def _resolve_unresolved(env: DaskDataFrameQueryBuilder, field: Bind) -> str:
    prop = env.model.flatprops.get(field.name)
    if prop:
        return prop.external.name
    else:
        raise PropertyNotFound(env.model, property=field.name)


@ufunc.resolver(DaskDataFrameQueryBuilder)
def count(env: DaskDataFrameQueryBuilder):
    return len(env.dataframe.index)


@ufunc.resolver(DaskDataFrameQueryBuilder, Expr)
def select(env: DaskDataFrameQueryBuilder, expr: Expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())

    if env.selected is not None:
        raise RuntimeError("`select` was already called.")

    env.selected = {}
    if args:
        for key, arg in args:
            env.selected[key] = env.call('select', arg)
    else:
        for prop in take(['_id', all], env.model.properties).values():
            if authorized(env.context, prop, Action.GETALL):
                env.selected[prop.place] = env.call('select', prop)


@ufunc.resolver(DaskDataFrameQueryBuilder, Bind)
def select(env: DaskDataFrameQueryBuilder, item: Bind, *, nested: bool = False):
    prop = _get_property_for_select(env, item.name, nested=nested)
    return env.call('select', prop)


def _get_property_for_select(
    env: DaskDataFrameQueryBuilder,
    name: str,
    *,
    nested: bool = False,
) -> Property:
    # TODO: `name` can refer to (in specified order):
    #       - var - a defined variable
    #       - param - a parameter if parametrization is used
    #       - item - an item of a dict or list
    #       - prop - a property
    #       Currently only `prop` is resolved.
    prop = env.model.flatprops.get(name)
    if prop and (
        # Check authorization only for top level properties in select list.
        # XXX: Not sure if nested is the right property to user, probably better
        #      option is to check if this call comes from a prepare context. But
        #      then how prepare context should be defined? Probably resolvers
        #      should be called with a different env class?
        #      tag:resolving_private_properties_in_prepare_context
        nested or
        authorized(env.context, prop, Action.SEARCH)
    ):
        return prop
    else:
        raise PropertyNotFound(env.model, property=name)


@ufunc.resolver(DaskDataFrameQueryBuilder, Property)
def select(env: DaskDataFrameQueryBuilder, prop: Property) -> Selected:
    if prop.place not in env.resolved:
        if isinstance(prop.external, list):
            raise SourceCannotBeList(prop)
        if prop.external.prepare is not NA:
            # If `prepare` formula is given, evaluate formula.
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            # XXX: Maybe interpretation of prepare formula should be done under
            #      a different Env? This way, select resolvers would know when
            #      properties are resolved under a formula context and for
            #      example would be able to decide to not check permissions on
            #      properties.
            #      tag:resolving_private_properties_in_prepare_context
            result = env.call('select', prop.dtype, result)
        elif prop.external and prop.external.name:
            # If prepare is not given, then take value from `source`.
            result = env.call('select', prop.dtype)
        elif prop.is_reserved():
            # Reserved properties never have external source.
            result = env.call('select', prop.dtype)
        else:
            # If `source` is not given, return None.
            result = Selected(prop=prop, prep=None)
        assert isinstance(result, Selected), prop
        env.resolved[prop.place] = result
    return env.resolved[prop.place]


@ufunc.resolver(DaskDataFrameQueryBuilder, DataType)
def select(env: DaskDataFrameQueryBuilder, dtype: DataType) -> Selected:
    return Selected(
        item=dtype.prop.external.name,
        prop=dtype.prop,
    )


@ufunc.resolver(DaskDataFrameQueryBuilder, DataType, object)
def select(env: DaskDataFrameQueryBuilder, dtype: DataType, prep: Any) -> Selected:
    if isinstance(prep, str):
        # XXX: Backwards compatibility thing.
        #      str values are interpreted as Bind values and Bind values are
        #      assumed to be properties. So here we skip
        #      `env.call('select', prep)` and return `prep` as is.
        #      This should be eventually removed, once backwards compatibility
        #      for resolving strings as properties is removed.
        return Selected(prop=dtype.prop, prep=prep)
    elif isinstance(prep, Expr):
        # If `prepare` expression returns another expression, then this means,
        # it must be processed on values returned by query.
        prop = dtype.prop
        if prop.external and prop.external.name:
            sel = env.call('select', dtype)
            return Selected(item=sel.item, prop=sel.prop, prep=prep)
        else:
            return Selected(item=None, prop=prop, prep=prep)
    else:
        if prep is not None:
            result = env.call('select', prep)
        else:
            result = None
        return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(DaskDataFrameQueryBuilder, PrimaryKey)
def select(
    env: DaskDataFrameQueryBuilder,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', prop)
    else:
        result = [
            env.call('select', prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(DaskDataFrameQueryBuilder, Ref, object)
def select(env: DaskDataFrameQueryBuilder, dtype: Ref, prep: Any) -> Selected:
    fpr = ForeignProperty(None, dtype, dtype.model.properties['_id'].dtype)
    return Selected(
        prop=dtype.prop,
        prep=env.call('select', fpr, fpr.right.prop),
    )


@ufunc.resolver(DaskDataFrameQueryBuilder, GetAttr)
def select(env: DaskDataFrameQueryBuilder, attr: GetAttr) -> Selected:
    """For things like select(foo.bar.baz)."""

    fpr: ForeignProperty = env.call('_resolve_getattr', attr)
    raise NotImplementedFeature(fpr.left.prop.model, feature="Ability to use foreign properties")
    return Selected(
        prop=fpr.right.prop,
        prep=env.call('select', fpr, fpr.right.prop),
    )


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
) -> Selected:
    return env.call('select', fpr, fpr.right.prop)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, Property)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    prop: Property,
) -> Selected:
    resolved_key = fpr / prop
    if resolved_key not in env.resolved:
        if isinstance(prop.external, list):
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external.prepare is not NA:

            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            result = env.call('select', fpr, prop.dtype, result)
        else:
            result = env.call('select', fpr, prop.dtype)
        assert isinstance(result, Selected), prop
        env.resolved[resolved_key] = result
    return env.resolved[resolved_key]


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, DataType)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
) -> Selected:
    # TODO need join for this to work
    return Selected(
        item=dtype.prop.name,
        prop=dtype.prop,
    )


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, DataType, object)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    prep: Any,
) -> Selected:
    result = env.call('select', fpr, prep)
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, tuple)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', fpr, v) for v in prep)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, Bind)
def select(env: DaskDataFrameQueryBuilder, fpr: ForeignProperty, item: Bind):
    model = fpr.right.prop.model
    prop = model.flatprops.get(item.name)
    if prop and authorized(env.context, prop, Action.SEARCH):
        return env.call('select', fpr, prop)
    else:
        raise PropertyNotFound(model, property=item.name)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', fpr, prop)
    else:
        result = [
            env.call('select', fpr, prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(DaskDataFrameQueryBuilder, list)
def select(
    env: DaskDataFrameQueryBuilder,
    prep: List[Any],
) -> List[Any]:
    return [env.call('select', v) for v in prep]


@ufunc.resolver(DaskDataFrameQueryBuilder, tuple)
def select(
    env: DaskDataFrameQueryBuilder,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', v) for v in prep)


@ufunc.resolver(DaskDataFrameQueryBuilder, dict)
def select(
    env: DaskDataFrameQueryBuilder,
    prep: Dict[str, Any],
) -> Dict[str, Any]:
    return {k: env.call('select', v) for k, v in prep.items()}



from typing import Any

from spinta.components import Property, Model
from spinta.core.ufuncs import Env
from spinta.datasets.enums import Level
from spinta.datasets.keymaps.components import KeyMap
from spinta.exceptions import GivenValueCountMissmatch
from spinta.types.datatype import Ref, Array


def handle_ref_key_assignment(keymap: KeyMap, env: Env, value: Any, ref: Ref) -> dict:
    keymap_name = ref.model.model_type()
    if ref.refprops != ref.model.external.pkeys:
        keymap_name = f'{keymap_name}.{"_".join(prop.name for prop in ref.refprops)}'

    if not isinstance(value, (tuple, list)):
        value = [value]
    else:
        value = list(value)

    prop_count_mapping = {}
    for prop in ref.refprops:
        if isinstance(prop.dtype, Array) and env:
            items = env.resolve(prop.external.prepare)
            prop_count_mapping[prop.name] = len(items)
        else:
            prop_count_mapping[prop.name] = 1
    expected_count = sum(item for item in prop_count_mapping.values())
    if len(value) != expected_count:
        raise GivenValueCountMissmatch(given_count=len(value), expected_count=expected_count)

    if not ref.prop.level or ref.prop.level.value > 3:
        if len(value) == 1:
            value = value[0]
        val = keymap.encode(keymap_name, value)
        val = {'_id': val}
    else:
        val = {}
        i = 0
        for prop, count in prop_count_mapping.items():
            values = value[i:i + count]
            if len(values) == 1:
                values = values[0]
            val[prop] = values
    return val


def generate_pk_for_row(model: Model, row: Any, keymap, pk_val: Any):
    pk = None
    if model.base and ((model.base.level and model.base.level >= Level.identifiable) or not model.base.level):
        pk_val_base = extract_values_from_row(row, model.base.parent, model.base.pk or model.base.parent.external.pkeys)
        key = model.base.parent.model_type()
        if model.base.pk and model.base.pk != model.base.parent.external.pkeys:
            joined = '_'.join(pk.name for pk in model.base.pk)
            key = f'{key}.{joined}'
        pk = keymap.encode(key, pk_val_base)

    pk = keymap.encode(model.model_type(), pk_val, pk)
    if pk and model.required_keymap_properties:
        for combination in model.required_keymap_properties:
            joined = '_'.join(combination)
            key = f'{model.model_type()}.{joined}'
            val = extract_values_from_row(row, model, combination)
            keymap.encode(key, val, pk)
    return pk


def extract_values_from_row(row: Any, model: Model, keys: list):
    return_list = []
    for key in keys:
        if isinstance(key, str):
            key = model.properties[key]
        if isinstance(key, Property):
            if key.external and key.external.name:
                key = key.external.name
            else:
                key = key.name

        return_list.append(row[key])
    if len(return_list) == 1:
        return_list = return_list[0]
    return return_list



from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.components import Context, Model
from spinta.types.datatype import Object
from spinta.migrations import SchemaVersion


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, Model, Object)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: Model,
    current: Object,
):
    for _, prop in current.properties.items():
        commands.freeze(context, version, backend, freezed, prop.dtype)


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, type(None), Object)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: type(None),
    current: Object,
):
    pr = []
    for name, prop in current.properties.items():
        pr += commands.freeze(context, version, backend, None, prop.dtype)
    return pr


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, Object, Object)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: Object,
    current: Object,
):
    for name in set(freezed.properties) | set(current.properties):
        freezed_dtype = freezed.prop.model
        if name in freezed.properties:
            freezed_dtype = freezed.properties[name].dtype

        current_dtype = current.prop.model
        if name in current.properties:
            current_dtype = current.properties[name].dtype

        commands.freeze(
            context,
            version,
            backend,
            freezed_dtype,
            current_dtype,
        )



import io
import json
import pathlib

import numpy as np
import urllib
from typing import Dict, Any, Iterator

import dask
import requests
from dask.dataframe import DataFrame
from dask.dataframe.utils import make_meta
from lxml import etree

from spinta import commands
from spinta.components import Context, Property, Model, UrlParams
from spinta.core.ufuncs import Expr
from spinta.datasets.backends.dataframe.components import DaskBackend, Csv, Xml, Json
from spinta.datasets.backends.dataframe.commands.query import DaskDataFrameQueryBuilder, Selected
from spinta.datasets.backends.dataframe.ufuncs.components import TabularResource

from spinta.datasets.backends.helpers import handle_ref_key_assignment
from spinta.datasets.helpers import get_enum_filters, get_ref_filters
from spinta.datasets.keymaps.components import KeyMap
from spinta.datasets.utils import iterparams
from spinta.dimensions.enum.helpers import get_prop_enum
from spinta.exceptions import ValueNotInEnum, PropertyNotFound, NoExternalName
from spinta.manifests.components import Manifest
from spinta.manifests.dict.helpers import is_list_of_dicts, is_blank_node
from spinta.types.datatype import PrimaryKey, Ref, DataType, Boolean, Number, Integer, DateTime, Time
from spinta.typing import ObjectData
from spinta.ufuncs.helpers import merge_formulas
from spinta.utils.data import take
from spinta.utils.schema import NA


@commands.load.register(Context, DaskBackend, dict)
def load(context: Context, backend: DaskBackend, config: Dict[str, Any]):
    pass


@commands.prepare.register(Context, DaskBackend, Manifest)
def prepare(context: Context, backend: DaskBackend, manifest: Manifest):
    pass


@commands.bootstrap.register(Context, DaskBackend)
def bootstrap(context: Context, backend: DaskBackend):
    pass


def _resolve_expr(
    context: Context,
    row: Dict[str, Any],
    sel: Selected,
) -> Any:
    if sel.item is None:
        val = None
    else:
        val = row[sel.item]
    raise NotImplementedError
    return val


def _get_row_value(context: Context, row: Dict[str, Any], sel: Any) -> Any:
    if isinstance(sel, Selected):
        if isinstance(sel.prep, Expr):
            val = _resolve_expr(context, row, sel)
        elif sel.prep is not NA:
            val = _get_row_value(context, row, sel.prep)
        else:
            if sel.item in row.keys():
                val = row[sel.item]
            else:
                raise PropertyNotFound(
                    sel.prop.model,
                    property=sel.prop.name,
                    external=sel.prop.external.name,
                )

        if enum := get_prop_enum(sel.prop):
            if val is None:
                pass
            elif str(val) in enum:
                item = enum[str(val)]
                if item.prepare is not NA:
                    val = item.prepare
            else:
                raise ValueNotInEnum(sel.prop, value=val)

        return val
    if isinstance(sel, tuple):
        return tuple(_get_row_value(context, row, v) for v in sel)
    if isinstance(sel, list):
        return [_get_row_value(context, row, v) for v in sel]
    if isinstance(sel, dict):
        return {k: _get_row_value(context, row, v) for k, v in sel.items()}
    return sel


def _select(prop: Property, df: DataFrame):
    if prop.external is None or not prop.external.name:
        raise NoExternalName(prop)
    column = prop.external.name
    if column not in df.columns:
        raise PropertyNotFound(
            prop.model,
            property=prop.name,
            external=prop.external.name,
        )
    return Selected(
        item=column,
        prop=prop,
    )


def _select_primary_key(model: Model, df: DataFrame):
    pkeys = model.external.pkeys

    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = _select(prop, df)
    else:
        result = [
            _select(prop, df)
            for prop in pkeys
        ]
    return Selected(prop=model.properties['_id'], prep=result)


def _parse_xml(data, source: str, model_props: Dict):
    elem_list = []
    added_root_elements = []
    for action, elem in etree.iterparse(data, events=('start', 'end'), remove_blank_text=True):
        if action == 'start':
            values = elem.xpath(source)

            # Check if it already has been added
            if not set(values).issubset(added_root_elements):
                for value in values:
                    if value not in added_root_elements:
                        new_dict = {}

                        # Go through each prop source with xpath of root path
                        for prop in model_props.values():
                            v = value.xpath(prop["source"])
                            new_value = None
                            if v:
                                v = v[0]
                                if isinstance(v, etree._Element):

                                    # Check if prop has pkeys (means its ref type)
                                    # If pkeys exists iterate with xpath and get the values
                                    pkeys = prop["pkeys"]

                                    if pkeys and prop["source"].endswith(('.', '/')):
                                        ref_keys = []
                                        for key in pkeys:
                                            ref_item = v.xpath(key)
                                            if ref_item:
                                                ref_item = ref_item[0]
                                                if isinstance(ref_item, etree._Element):
                                                    if ref_item.text:
                                                        ref_item = ref_item.text
                                                    elif ref_item.attrib:
                                                        ref_item = ref_item.attrib
                                                    else:
                                                        ref_item = None
                                                elif not ref_item:
                                                    ref_item = None
                                                ref_keys.append(str(ref_item))
                                            if len(ref_keys) == 1:
                                                new_value = ref_keys[0]
                                            else:
                                                new_value = ref_keys
                                    else:
                                        if v.text:
                                            new_value = str(v.text)
                                        elif v.attrib:
                                            new_value = v.attrib.values()
                                        else:
                                            new_value = None
                                elif v:
                                    new_value = str(v)
                                else:
                                    new_value = None
                            new_dict[prop["source"]] = new_value
                        elem_list.append(new_dict)
                        added_root_elements.append(value)

            # Required for optimization
            elem.clear()
    return elem_list


def _parse_json(data, source: str, model_props: Dict):
    data = json.loads(data)
    source_list = source.split(".")
    # Prepare source for blank nodes
    blank_nodes = is_blank_node(data)
    if source == ".":
        source_list = ["."]
    else:
        if blank_nodes:
            source_list = ["."] + source_list
    if blank_nodes:
        for prop in model_props.values():
            if prop["root_source"][0] != '.':
                prop["root_source"].insert(0, '.')

    values = _parse_json_with_params(data, source_list, model_props)
    return values


def _parse_json_with_params(data, source: list, model_props: dict):
    result = []

    def get_prop_value(items, pkeys: list, prop_source: list):
        new_value = items

        # Remove empty values from split()
        prop_source = [prop for prop in prop_source if prop]
        if pkeys:
            ref_keys = []
            for key in pkeys:
                split = key.split(".")
                split = [prop for prop in split if prop]
                new_value = items
                for id, prop in enumerate(split):
                    if isinstance(new_value, dict):
                        if prop in new_value.keys():
                            new_value = new_value[prop]
                        else:
                            break
                    else:
                        break
                    if id == len(split) - 1:
                        ref_keys.append(new_value)
            if len(ref_keys) == 1:
                return ref_keys[0]
            else:
                return ref_keys
        else:
            for id, prop in enumerate(prop_source):
                if isinstance(new_value, dict):
                    if prop in new_value.keys():
                        new_value = new_value[prop]
                    else:
                        return None
                else:
                    return None
                if id == len(prop_source) - 1:
                    return new_value

    def traverse_json(items, current_value: Dict, current_path: int = 0, in_model: bool = False):
        last_path = source[:current_path + 1]
        c_path = source[current_path]
        if c_path.endswith("[]"):
            c_path = c_path[:-2]
        if isinstance(items, dict):
            if in_model or c_path == ".":
                item = items
            else:
                if c_path in items.keys():
                    item = items[c_path]
                else:
                    return
            if isinstance(item, list) and is_list_of_dicts(item):
                traverse_json(item, current_value, current_path, True)
            else:
                for prop in model_props.values():
                    prop_path = prop["root_source"]
                    this_path = last_path
                    if prop_path == this_path:
                        current_value[prop["source"]] = get_prop_value(item, prop["pkeys"], prop["source"].split("."))
                if current_path < len(source) - 1:
                    traverse_json(item, current_value, current_path + 1, False)
                else:
                    result.append(current_value)
        elif isinstance(items, list) and is_list_of_dicts(items):
            for item in items:
                traverse_json(item, current_value.copy(), current_path, in_model)

    c_value = {}
    for prop in model_props.values():
        c_value[prop["source"]] = None
    traverse_json(data, current_value=c_value)
    return result


def _get_data_xml(url: str, source: str, model_props: Dict):
    if url.startswith(('http', 'https')):
        f = requests.get(url)
        return _parse_xml(io.BytesIO(f.content), source, model_props)
    else:
        with open(url, 'rb') as f:
            return _parse_xml(f, source, model_props)


def _get_data_json(url: str, source: str, model_props: Dict):
    if url.startswith(('http', 'https')):
        f = requests.get(url)
        return _parse_json(f.text, source, model_props)
    else:
        with pathlib.Path(url).open(encoding='utf-8-sig') as f:
            return _parse_json(f.read(), source, model_props)


def _get_prop_full_source(source: str, prop_source: str):
    if prop_source.startswith(".."):
        split = prop_source[1:].count(".")
        split_source = source.split(".")
        value = f'{".".join(split_source[:len(split_source) - split])}'
        if value == '':
            return '.'
        return value
    else:
        return source


def _get_pkeys_if_ref(prop: Property):
    return_list = []
    if isinstance(prop.dtype, Ref):
        for key in prop.dtype.refprops:
            if key.external.name:
                return_list.append(key.external.name)
    return return_list


def _get_dask_dataframe_meta(model: Model):
    dask_meta = {}
    for prop in model.properties.values():
        if prop.external and prop.external.name:
            dask_meta[prop.external.name] = spinta_to_np_dtype(prop.dtype)
    return dask_meta


@commands.getall.register(Context, Model, Json)
def getall(
    context: Context,
    model: Model,
    backend: Json,
    *,
    query: Expr = None,
    **kwargs
) -> Iterator[ObjectData]:
    base = model.external.resource.external
    builder = DaskDataFrameQueryBuilder(context)
    builder.update(model=model)
    props = {}
    for prop in model.properties.values():
        if prop.external and prop.external.name:
            root_source = _get_prop_full_source(model.external.name, prop.external.name)
            props[prop.external.name] = {
                "source": prop.external.name,
                "root_source": root_source.split(".") if root_source != "." else ["."],
                "pkeys": _get_pkeys_if_ref(prop)
            }

    for params in iterparams(context, model):
        meta = _get_dask_dataframe_meta(model)
        df = dask.bag.from_sequence([base]).map(
            _get_data_json,
            source=model.external.name,
            model_props=props
        ).flatten().to_dataframe(meta=meta)
        yield from _dask_get_all(context, query, df, backend, model, builder)


@commands.getall.register(Context, Model, Xml)
def getall(
    context: Context,
    model: Model,
    backend: Xml,
    *,
    query: Expr = None,
    **kwargs
) -> Iterator[ObjectData]:
    base = model.external.resource.external
    builder = DaskDataFrameQueryBuilder(context)
    builder.update(model=model)
    props = {}
    for prop in model.properties.values():
        if prop.external and prop.external.name:
            props[prop.name] = {
                "source": prop.external.name,
                "pkeys": _get_pkeys_if_ref(prop)
            }

    for params in iterparams(context, model):
        df = dask.bag.from_sequence([base]).map(
            _get_data_xml,
            source=model.external.name,
            model_props=props
        ).flatten().to_dataframe(meta=_get_dask_dataframe_meta(model))
        yield from _dask_get_all(context, query, df, backend, model, builder)


@commands.getall.register(Context, Model, Csv)
def getall(
    context: Context,
    model: Model,
    backend: Csv,
    *,
    query: Expr = None,
    **kwargs
) -> Iterator[ObjectData]:
    base = model.external.resource.external
    resource_builder = TabularResource(context)
    resource_builder.resolve(model.external.resource.prepare)

    builder = DaskDataFrameQueryBuilder(context)
    builder.update(model=model)
    for params in iterparams(context, model):
        if base:
            url = urllib.parse.urljoin(base, model.external.name)
        else:
            url = model.external.name
        url = url.format(**params)
        df = dask.dataframe.read_csv(url, sep=resource_builder.seperator)
        yield from _dask_get_all(context, query, df, backend, model, builder)


def _dask_get_all(context: Context, query: Expr, df: dask.dataframe, backend: DaskBackend, model: Model, env: DaskDataFrameQueryBuilder):
    keymap: KeyMap = context.get(f'keymap.{model.keymap.name}')

    query = merge_formulas(model.external.prepare, query)
    query = merge_formulas(query, get_enum_filters(context, model))
    query = merge_formulas(query, get_ref_filters(context, model))
    env = env.init(backend, df)
    expr = env.resolve(query)
    where = env.execute(expr)
    qry = env.build(where)
    for i, row in qry.iterrows():
        row = row.to_dict()
        res = {
            '_type': model.model_type(),
        }
        for key, sel in env.selected.items():
            val = _get_row_value(context, row, sel)
            if sel.prop:
                if isinstance(sel.prop.dtype, PrimaryKey):
                    val = keymap.encode(sel.prop.model.model_type(), val)
                elif isinstance(sel.prop.dtype, Ref):
                    val = handle_ref_key_assignment(keymap, env, val, sel.prop.dtype)
            res[key] = val
        res = commands.cast_backend_to_python(context, model, backend, res)
        yield res


@commands.wait.register(Context, DaskBackend)
def wait(context: Context, backend: DaskBackend, *, fail: bool = False) -> bool:
    return True


@commands.spinta_to_np_dtype.register(Boolean)
def spinta_to_np_dtype(dtype: Boolean):
    return "boolean"


@commands.spinta_to_np_dtype.register(Number)
def spinta_to_np_dtype(dtype: Number):
    return np.dtype("f")


@commands.spinta_to_np_dtype.register(Integer)
def spinta_to_np_dtype(dtype: Integer):
    return "Int64"


@commands.spinta_to_np_dtype.register(DateTime)
def spinta_to_np_dtype(dtype: DateTime):
    return "datetime64[ns]"


@commands.spinta_to_np_dtype.register(DataType)
def spinta_to_np_dtype(dtype: DataType):
    return np.dtype("S")



from typing import List

from spinta import commands
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name as gcn
from spinta.components import Property
from spinta.datasets.enums import Level
from spinta.exceptions import PropertyNotFound
from spinta.types.datatype import DataType, Ref, BackRef, String
import sqlalchemy as sa

from spinta.types.text.components import Text


def convert_str_to_columns(table: sa.Table, prop: Property, columns: List[str]) -> List[sa.Column]:
    return_list = []
    for column in columns:
        return_list.append(convert_str_to_column(table, prop, column))
    return return_list


def convert_str_to_column(table: sa.Table, prop: Property, column: str) -> sa.Column:
    if column in table.c:
        return table.c[column]
    else:
        raise PropertyNotFound(prop.dtype)


def get_table(backend: PostgreSQL, prop: Property) -> sa.Table:
    if prop.list:
        table = backend.get_table(prop.list, TableType.LIST)
    else:
        table = backend.get_table(prop)
    return table


@commands.get_column.register(PostgreSQL, Property)
def get_column(backend: PostgreSQL, prop: Property, **kwargs):
    return commands.get_column(backend, prop.dtype, **kwargs)


@commands.get_column.register(PostgreSQL, DataType)
def get_column(backend: PostgreSQL, dtype: DataType, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    if table is None:
        table = get_table(backend, prop)
    column_name = gcn(dtype.prop)
    return convert_str_to_column(table, prop, column_name)


@commands.get_column.register(PostgreSQL, String)
def get_column(backend: PostgreSQL, dtype: String, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    parent = prop.parent

    if table is None:
        table = get_table(backend, prop)
    column_name = gcn(parent if parent and isinstance(parent.dtype, Text) else dtype.prop)
    column = convert_str_to_column(table, prop, column_name)

    if parent and isinstance(parent.dtype, Text):
        column = column[dtype.prop.name].label(dtype.prop.place.replace('@', '.'))
    return column


@commands.get_column.register(PostgreSQL, Text)
def get_column(backend: PostgreSQL, dtype: Text, table: sa.Table = None, langs: list = [], default_langs: list = [], push: bool = False, **kwargs):
    prop = dtype.prop
    column_name = gcn(dtype.prop)
    if table is None:
        table = get_table(backend, prop)

    if push:
        column = convert_str_to_column(table, prop, column_name)
        return column

    existing_langs = list(dtype.langs.keys())
    lang_prop = None
    if langs:
        for lang in langs:
            if lang in dtype.langs:
                lang_prop = dtype.langs[lang]
                break
    if not lang_prop and default_langs:
        for lang in default_langs:
            if lang in dtype.langs:
                lang_prop = dtype.langs[lang]
                break
    if not lang_prop:
        lang_prop = dtype.langs[existing_langs[0]]
    column = convert_str_to_column(table, prop, column_name)
    column = column[lang_prop.name].label(prop.place)
    return column


@commands.get_column.register(PostgreSQL, Ref)
def get_column(backend: PostgreSQL, dtype: Ref, table: sa.Table = None, **kwargs):
    column = gcn(dtype.prop)
    prop = dtype.prop
    columns = []
    if table is None:
        table = get_table(backend, prop)
    if prop.level is None or prop.level > Level.open:
        if prop.list is not None:
            column = '_id'
        else:
            column += '._id'
        return convert_str_to_column(table, prop, column)
    else:
        for refprop in prop.dtype.refprops:
            if prop.list is not None:
                columns.append(refprop.name)
            else:
                columns.append(f'{column}.{refprop.name}')
        return convert_str_to_columns(table, prop, columns)


@commands.get_column.register(PostgreSQL, BackRef)
def get_column(backend: PostgreSQL, dtype: BackRef, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    r_prop = dtype.refprop
    columns = []
    if table is None:
        table = get_table(backend, dtype.prop)
    if r_prop.level is None or r_prop.level > Level.open:
        column = '_id'
        return convert_str_to_column(table, prop, column)
    else:
        for refprop in r_prop.dtype.refprops:
            columns.append(refprop.name)
        return convert_str_to_columns(table, prop, columns)



from typing import List

from spinta import commands
from spinta.components import Context
from spinta.utils.scopes import name_to_scope


@commands.get_model_scopes.register(Context, str, list)
def get_model_scopes(context: Context, model: str, actions: List[str]):
    config = context.get('config')
    return [
        name_to_scope('{prefix}{name}_{action}', model, maxlen=config.scope_max_length, params={
            'prefix': config.scope_prefix,
            'action': action,
        })
        for action in actions
    ]



import enum
import itertools
from typing import Any, Tuple, Dict, List

from rdflib import Graph
from rdflib.query import Result
from rdflib.term import URIRef

from spinta.manifests.rdf.components import RdfManifest
from spinta.manifests.tabular.helpers import get_relative_model_name
from spinta.utils.naming import to_model_name, to_property_name

QUERY = '''
    select distinct
        ?schema
        ?base
        ?model
        ?property
        ?type
        ?title
        ?description
        ?order
    where {
        {
            ?model_shape a sh:NodeShape ;
                sh:targetClass ?model ;
                sh:property ?shape .

            ?shape sh:path ?property .

            optional { ?shape sh:datatype ?type }
            optional { ?shape sh:property ?type }
            optional { ?shape sh:class ?type }
            optional { ?shape sh:name ?title }
            optional { ?shape sh:description ?description }
            optional { ?shape sh:order ?order }

            bind("shacl" as ?schema)
        }

        union

        {
            ?shape a sh:NodeShape ;
                sh:targetClass ?model .

            optional { ?shape sh:name ?title }
            optional { ?shape sh:description ?description }

            bind("shacl" as ?schema)
        }

        union

        {
            ?property a rdf:Property .
            optional { ?property rdfs:domain ?model }
            optional { ?property rdfs:range ?type }
            optional { ?property rdfs:label ?title filter langMatches(lang(?title), "EN") }
            optional { ?property rdfs:comment ?description filter langMatches(lang(?description), "EN") }
            bind("rdfs" as ?schema)
        }

        union

        {
            ?property a rdf:Property .
            optional { ?property rdfs:isDefinedBy ?description }
            bind("rdfs" as ?schema)
        }
        
        union
        {
            ?property a owl:ObjectProperty .
            optional { ?property rdfs:label ?title }
            optional { ?property rdfs:comment ?description }
            optional { ?property rdfs:domain ?type }
            bind("owl" as ?schema)
        }

        union

        {
            ?property rdfs:domain ?model .
            optional { ?property rdfs:range ?type }
            optional { ?property rdfs:label ?title filter langMatches(lang(?title), "EN") }
            optional { ?property rdfs:comment ?description filter langMatches(lang(?description), "EN") }
            bind("rdfs" as ?schema)
        }

        union

        {
            ?property rdfs:domain ?model .
            optional { ?property rdfs:isDefinedBy ?description }
            bind("rdfs" as ?schema)
        }

        union

        {
            ?property rdfs:range ?type .
            optional { ?property rdfs:domain ?model }
            optional { ?property rdfs:label ?title filter langMatches(lang(?title), "EN") }
            optional { ?property rdfs:comment ?description filter langMatches(lang(?description), "EN") }
            bind("rdfs" as ?schema)
        }

        union

        {
            ?property rdfs:range ?type .
            optional { ?property rdfs:isDefinedBy ?description }
            bind("rdfs" as ?schema)
        }

        union

        {
            ?model a rdfs:Class .
            optional { ?model rdfs:label ?title filter langMatches(lang(?title), "EN") }
            optional { ?model rdfs:comment ?description filter langMatches(lang(?description), "EN") }
            bind("rdfs" as ?schema)
        }

        union

        {
            ?model a rdfs:Class .
            optional { ?model rdfs:isDefinedBy ?description }
            bind("rdfs" as ?schema)
        }
        
        union
        
        {
            ?model a owl:Class .
            optional { ?model rdfs:label ?title }
            optional { ?model rdfs:comment ?description }
            bind("owl" as ?schema)
        }

        union

        {
            ?p rdfs:domain ?model .
            optional { ?model rdfs:label ?title filter langMatches(lang(?title), "EN") }
            optional { ?model rdfs:comment ?description filter langMatches(lang(?description), "EN") }
            bind("rdfs" as ?schema)
        }

        union

        {
            ?p rdfs:domain ?model .
            optional { ?model rdfs:isDefinedBy ?description }
            bind("rdfs" as ?schema)
        }

        union

        {
            ?model rdfs:subClassOf ?base .
            bind("rdfs" as ?schema)
        }

    }
    '''


class Schema(enum.IntEnum):
    undefined = 0
    rdfs = 1
    owl = 2
    shacl = 3


def read_rdf_manifest(
    manifest: RdfManifest,
    lang: str = 'en',
):
    g = Graph(bind_namespaces='rdflib')
    g.parse(manifest.path, format=manifest.format)

    schemas = []
    dataset = {
        'type': 'dataset',
        'name': manifest.path.split('/')[-1].rsplit('.', 1)[0],
        'prefixes': {}
    }
    for i, (name, url) in enumerate(g.namespaces()):
        dataset['prefixes'][name] = {
            'type': 'prefix',
            'eid': i,
            'name': name,
            'uri': url,
        }
    schemas.append(dataset)

    models = _prepare_data(g, g.query(QUERY), dataset, lang)
    schemas.extend(models.values())

    yield from _get_schemas(
        g,
        schemas,
        dataset,
        list(models.keys())
    )


def _get_schemas(
    graph: Graph,
    schemas: List[Dict],
    dataset: Dict,
    models: List,
):
    for i, row in enumerate(schemas, 1):

        if 'schema' in row:
            row.pop('schema')

        if 'description' in row:
            row['description'] = '\n'.join(sorted(set(row['description'])))

        if 'properties' in row:
            for prop in row['properties'].values():
                if 'schema' in prop:
                    prop.pop('schema')

                prop['description'] = '\n'.join(sorted(set(prop['description'])))
                prop['type'], ref_model = _to_type_name(graph, dataset, models, prop['type'])
                if ref_model:
                    prop['model'] = ref_model

        yield i, row


def _prepare_data(
    graph: Graph,
    query: Result,
    dataset: Dict,
    lang: str = 'en'
) -> Dict:
    data = {}
    for (model, prop), rows in itertools.groupby(query, key=_get_key):
        dim = _get_dimension(model, prop)

        for row in rows:
            if dim == 'model':
                _read_model(graph, dataset, data, row, lang)

            elif dim == 'property':
                _read_property(graph, dataset, data, row, lang)

    return data


def _read_model(
    graph: Graph,
    dataset: Dict,
    data: Dict,
    row: Dict,
    lang: str = 'en'
) -> None:
    schema = _load_schema(str(row['schema']))
    model, uri = _to_model_name(graph, dataset, row['model'])
    base, _ = _to_model_name(graph, dataset, row['base'])
    title = _get_value_for_language(row['title'], lang)
    properties = {}
    descriptions = []

    if description := _get_value_for_language(row['description'], lang):
        descriptions.append(description)

    if model in data:
        existing_model = data[model]
        properties = existing_model['properties']
        title = title or existing_model['title']
        descriptions.extend(existing_model['description'])
        if not base and existing_model['base']:
            base = existing_model['base']['parent']

        if data[model]['schema'] > schema:
            schema = existing_model['schema']
            base = existing_model['base']['parent'] if existing_model['base'] else base
            title = existing_model['title'] or title

    if base and base not in data:
        # Add temporary model for base
        data.update({
            base: {
                'type': 'model',
                'schema': Schema.undefined,
                'name': base,
                'base': None,
                'properties': {},
                'uri': "",
                'access': 'open',
                'description': [],
                'title': '',
                'external': {
                    'dataset': dataset['name'],
                },
            }
        })

    data.update({
        model: {
            'type': 'model',
            'schema': schema,
            'name': model,
            'base': {
                'name': base.split('/')[-1],
                'parent': base,
            } if base else None,
            'title': title,
            'description': descriptions,
            'properties': properties,
            'uri': uri,
            'access': 'open',
            'external': {
                'dataset': dataset['name'],
            },
        }
    })


def _read_property(
    graph: Graph,
    dataset: Dict,
    data: Dict,
    row: Dict,
    lang: str = 'en'
) -> None:
    schema = _load_schema(str(row['schema']))
    model, model_uri = _to_model_name(graph, dataset, row['model'] or 'rdfs:Resource')
    prop, prop_uri = _to_property_name(graph, row['property'])
    type = _parse_uri(graph, row['type'])
    title = _get_value_for_language(row['title'], lang)
    descriptions = []

    if description := _get_value_for_language(row['description'], lang):
        descriptions.append(description)

    if model not in data:
        # Add temporary model for property
        data.update({
            model: {
                'type': 'model',
                'schema': Schema.undefined,
                'name': model,
                'base': None,
                'properties': {},
                'uri': model_uri,
                'access': 'open',
                'description': [],
                'title': '',
                'external': {
                    'dataset': dataset['name'],
                },
            }
        })

    model = data[model]

    if prop in model['properties']:
        existing_prop = model['properties'][prop]
        title = title or existing_prop['title']
        type = type or existing_prop['type']
        descriptions.extend(existing_prop['description'])

        if model['properties'][prop]['schema'] > schema:
            schema = existing_prop['schema']
            type = existing_prop['type'] or type
            title = existing_prop['title'] or title

    model['properties'].update({
        prop: {
            'schema': schema,
            'type': type,
            'title': title,
            'description': descriptions,
            'uri': prop_uri,
            'access': 'open',
        }
    })


def _get_key(row: Dict) -> Tuple[str, str]:
    return row['model'], row['property']


def _get_dimension(model: str, prop: str) -> str:
    if model and not prop:
        return 'model'
    elif prop:
        return 'property'
    return ''


def _load_schema(schema: str) -> Schema:
    if schema == 'rdfs':
        return Schema.rdfs
    elif schema == 'owl':
        return Schema.owl
    elif schema == 'shacl':
        return Schema.shacl
    else:
        return Schema.undefined


def _to_model_name(
    graph: Graph,
    dataset: Dict,
    name: str
) -> Tuple[str, str]:
    name = _parse_uri(graph, name)
    uri = ''
    if name:
        if ':' in name:
            uri = name
            name = to_model_name(name.split(':', 1)[1])
        name = get_relative_model_name(dataset, name)
    return name, uri


def _to_property_name(graph: Graph, name: str) -> Tuple[str, str]:
    name = _parse_uri(graph, name)
    uri = ''
    if name and ':' in name:
        uri = name
        name = to_property_name(name.split(':', 1)[1])
    return name, uri


TYPES = {
    'xsd:hexBinary': 'binary',
    'xsd:duration': 'string',
    'xsd:decimal': 'number',
    'xsd:dateTime': 'datetime',
    'xsd:nonNegativeInteger': 'integer',
    'xsd:date': 'date',
    'xsd:dateTimeStamp': 'datetime',
    'xsd:gYear': 'string',
    'xsd:gYearMonth': 'string',
    'xsd:string': 'string',
    'dct:MediaType': 'file',
    'rdfs:Literal': 'string',
    'vcard:Kind': 'string',
    'foaf:Document': 'file',
    'dcterms:Location': 'string',
    'owl:Thing': 'string',
}


def _to_type_name(
    graph: Graph,
    dataset: Dict,
    models: List,
    name: str,
) -> Tuple[str, str]:
    ref_model = None
    type = 'string'
    if name:
        if name in TYPES:
            type = TYPES[name]
        else:
            model_name, _ = _to_model_name(graph, dataset, name)
            if model_name in models:
                type = 'ref'
                ref_model = model_name
    return type, ref_model


def _parse_uri(graph: Graph, name: Any) -> str:
    if isinstance(name, URIRef):
        return graph.qname(name)
    elif name:
        return str(name)
    else:
        return name


def _get_value_for_language(
    value: Any,
    lang: str
) -> str:
    if (
        value and
        hasattr(value, 'language') and
        value.language == lang
    ):
        return str(value)
    else:
        return ''



from typing import Dict, Any

import sqlalchemy as sa

from spinta import commands
from spinta.components import Context
from spinta.backends.postgresql.components import PostgreSQL


@commands.load.register(Context, PostgreSQL, dict)
def load(context: Context, backend: PostgreSQL, config: Dict[str, Any]):
    backend.dsn = config['dsn']
    backend.engine = sa.create_engine(backend.dsn, echo=False)
    backend.schema = sa.MetaData(backend.engine)
    backend.tables = {}


@commands.unload_backend.register()
def unload_backend(context: Context, backend: PostgreSQL):
    # Make sure all connections are released, since next test will create
    # another connection pool and connection pool is not reused between
    # tests. Maybe it would be a good idea to reuse same connection between
    # all tests?
    backend.engine.dispose()



from __future__ import annotations

import os
import shutil
import uuid
from typer import echo
from typing import Set
from typing import Type
from typing import Union, List, Tuple

import datetime
import enum
import json
import logging
import time
import pathlib
import base64

import ruamel.yaml

from starlette.responses import JSONResponse
from starlette.exceptions import HTTPException
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa

from authlib.jose import jwk
from authlib.jose import jwt
from authlib.jose.errors import JoseError
from authlib.oauth2 import OAuth2Request
from authlib.oauth2 import rfc6749
from authlib.oauth2 import rfc6750
from authlib.oauth2.rfc6749 import grants
from authlib.oauth2.rfc6750.errors import InsufficientScopeError
from authlib.oauth2.rfc6749.errors import InvalidClientError
from authlib.oauth2 import OAuth2Error

from spinta.components import Config
from spinta.components import ScopeFormatterFunc
from spinta.core.enums import Access
from spinta.components import Context, Action, Namespace, Model, Property
from spinta.exceptions import InvalidToken, NoTokenValidationKey, ClientWithNameAlreadyExists, ClientAlreadyExists
from spinta.exceptions import AuthorizedClientsOnly
from spinta.exceptions import BasicAuthRequired
from spinta.utils import passwords
from spinta.utils.scopes import name_to_scope
from spinta.utils.types import is_str_uuid

log = logging.getLogger(__name__)
yaml = ruamel.yaml.YAML(typ="safe")

yml = ruamel.yaml.YAML()
yml.indent(mapping=2, sequence=4, offset=2)
yml.width = 80
yml.explicit_start = False


class AuthorizationServer(rfc6749.AuthorizationServer):

    def __init__(self, context):
        super().__init__(
            query_client=self._query_client,
            generate_token=rfc6750.BearerToken(
                access_token_generator=self._generate_token,
                expires_generator=self._get_expires_in,
            ),
            save_token=self._save_token,
        )
        self.register_grant(grants.ClientCredentialsGrant)
        self._context = context
        self._private_key = load_key(context, KeyType.private, required=False)

    def enabled(self):
        return self._private_key is not None

    def create_oauth2_request(self, request):
        return get_auth_request(request)

    def handle_response(self, status_code, payload, headers) -> JSONResponse:
        return JSONResponse(payload, status_code=status_code, headers=dict(headers))

    def handle_error_response(
        self,
        request: OAuth2Request,
        error: OAuth2Error,
    ) -> JSONResponse:
        log.exception('Authorization server error: %s', error)
        return super().handle_error_response(request, error)

    def send_signal(self, *args, **kwargs):
        pass

    def _query_client(self, client_name):
        path = get_clients_path(self._context.get('config'))
        return query_client(path, client_name, is_name=True)

    def _save_token(self, token, request):
        pass

    def _get_expires_in(self, client, grant_type):
        return int(datetime.timedelta(days=10).total_seconds())

    def _generate_token(self, client: Client, grant_type, user, scope, **kwargs):
        expires_in = self._get_expires_in(client, grant_type)
        scopes = scope.split() if scope else []
        return create_access_token(self._context, self._private_key, client.id, expires_in, scopes)


class ResourceProtector(rfc6749.ResourceProtector):

    def __init__(
        self,
        context: Context,
        Validator: Type[rfc6750.BearerTokenValidator],
    ):
        self.TOKEN_VALIDATORS = {
            Validator.TOKEN_TYPE: Validator(context),
        }


class BearerTokenValidator(rfc6750.BearerTokenValidator):

    def __init__(self, context):
        super().__init__()
        self._context = context
        self._public_key = load_key(context, KeyType.public)

    def authenticate_token(self, token_string: str):
        return Token(token_string, self)

    def request_invalid(self, request):
        return False

    def token_revoked(self, token):
        return False


class Client(rfc6749.ClientMixin):
    id: str
    name: str
    secret_hash: str
    scopes: Set[str]

    def __init__(self, *, id_: str, name_: str, secret_hash: str, scopes: List[str]):
        self.id = id_
        self.name = name_
        self.secret_hash = secret_hash
        self.scopes = set(scopes)

    def __repr__(self):
        cls = type(self)
        return f'{cls.__module__}.{cls.__name__}(id={self.id!r})'

    def check_client_secret(self, client_secret):
        log.debug(f"Incorrect client {self.id!r} secret hash.")
        return passwords.verify(client_secret, self.secret_hash)

    def check_token_endpoint_auth_method(self, method: str):
        return method == 'client_secret_basic'

    def check_grant_type(self, grant_type: str):
        return grant_type == 'client_credentials'

    def check_requested_scopes(self, scopes: set):
        unknown_scopes = scopes - self.scopes
        if unknown_scopes:
            log.warning(f"requested unknown scopes: %s", ', '.join(sorted(unknown_scopes)))
            return False
        else:
            return True


class Token(rfc6749.TokenMixin):

    def __init__(self, token_string, validator: BearerTokenValidator):
        try:
            self._token = jwt.decode(token_string, validator._public_key)
        except JoseError as e:
            raise InvalidToken(error=str(e))

        self._validator = validator

    def valid_scope(self, scope, *, operator='AND'):
        if self._validator.scope_insufficient(self, scope, operator):
            return False
        else:
            return True

    def check_scope(self, scope, *, operator='AND'):
        if not self.valid_scope(scope, operator=operator):
            client_id = self._token['aud']

            if isinstance(scope, str):
                scope = [scope]

            missing_scopes = ', '.join(sorted(scope))

            # FIXME: this should be wrapped into UserError.
            if operator == 'AND':
                log.error(f"client {client_id!r} is missing required scopes: %s", missing_scopes)
                raise InsufficientScopeError(description=f"Missing scopes: {missing_scopes}")
            elif operator == 'OR':
                log.error(f"client {client_id!r} is missing one of required scopes: %s", missing_scopes)
                raise InsufficientScopeError(description=f"Missing one of scopes: {missing_scopes}")
            else:
                raise Exception(f"Unknown operator {operator}.")

    def get_expires_at(self):
        return self._token['exp']

    def get_scope(self):
        return self._token.get('scope', '')

    def get_sub(self):  # User.
        return self._token.get('sub', '')

    def get_aud(self):  # Client.
        return self._token.get('aud', '')

    def get_client_id(self):
        return self.get_aud()


class AdminToken(rfc6749.TokenMixin):

    def valid_scope(self, scope, **kwargs):
        return True

    def check_scope(self, scope, **kwargs):
        pass

    def get_sub(self):  # User.
        return 'admin'

    def get_aud(self):  # Client.
        return 'admin'

    def get_client_id(self):
        return self.get_aud()


def get_auth_token(context: Context) -> Token:
    scope = None  # Scopes will be validated later using Token.check_scope
    request: OAuth2Request = context.get('auth.request')

    config = context.get('config')
    if config.default_auth_client and 'authorization' not in request.headers:
        default_id = get_client_id_from_name(get_clients_path(config), config.default_auth_client)
        if default_id:
            token = create_client_access_token(context, default_id)
            request.headers = request.headers.mutablecopy()
            request.headers['authorization'] = f'Bearer {token}'

    elif config.http_basic_auth:
        token = get_token_from_http_basic_auth(context, request)
        if token is not None:
            request.headers = request.headers.mutablecopy()
            request.headers['authorization'] = f'Bearer {token}'

    resource_protector = context.get('auth.resource_protector')
    try:
        token = resource_protector.validate_request(scope, request)
    except JoseError as e:
        raise HTTPException(status_code=400, detail=e.error)
    return token


def get_token_from_http_basic_auth(context: Context, request: OAuth2Request):
    if 'authorization' not in request.headers:
        raise BasicAuthRequired()

    auth = request.headers['authorization']
    if ' ' not in auth:
        raise BasicAuthRequired()

    method, value = request.headers['authorization'].split(None, 1)
    method = method.lower()
    if method != 'basic':
        # Pass authentication to authlib.
        return

    value = base64.b64decode(value).decode()
    client, secret = value.split(':', 1)

    # Get client.
    try:
        config = context.get('config')
        client = query_client(get_clients_path(config), client, is_name=True)
    except InvalidClientError:
        raise BasicAuthRequired()

    # Check secret.
    if not client.check_client_secret(secret):
        raise BasicAuthRequired()

    return create_client_access_token(context, client)


def get_auth_request(request: dict) -> OAuth2Request:
    return OAuth2Request(
        request['method'],
        request['url'],
        request['body'],
        request['headers'],
    )


def create_key_pair():
    return rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())


class KeyType(enum.Enum):
    public = 'public'
    private = 'private'


def load_key(context: Context, key_type: KeyType, *, required: bool = True):
    key = None
    config = context.get('config')

    # Public key can be set via configuration.
    if key_type == KeyType.public:
        key = config.token_validation_key

    # Load key from a file.
    if key is None:
        keypath = config.config_path / 'keys' / f'{key_type.value}.json'
        if keypath.exists():
            with keypath.open() as f:
                key = json.load(f)

    if key is None:
        if required:
            raise NoTokenValidationKey(key_type=key_type.value)
        else:
            return

    if isinstance(key, dict) and 'keys' in key:
        # XXX: Maybe I should load all keys and then pick right one by algorithm
        #      used in token?
        keys = [k for k in key['keys'] if k['alg'] == 'RS512']
        key = keys[0]

    key = jwk.loads(key)
    return key


def create_client_access_token(context: Context, client: Union[str, Client]):
    private_key = load_key(context, KeyType.private)
    if isinstance(client, str):
        client = query_client(get_clients_path(context.get('config')), client)
    expires_in = int(datetime.timedelta(days=10).total_seconds())
    return create_access_token(context, private_key, client.id, expires_in, client.scopes)


def create_access_token(
    context: Context,
    private_key,
    client: str,
    expires_in: int = None,
    scopes: Set[str] = None,
):
    config = context.get('config')

    if expires_in is None:
        expires_in = int(datetime.timedelta(minutes=10).total_seconds())

    header = {
        'typ': 'JWT',
        'alg': 'RS512',
    }

    iat = int(time.time())
    exp = iat + expires_in
    scopes = ' '.join(sorted(scopes)) if scopes else ''
    payload = {
        'iss': config.server_url,
        'sub': client,
        'aud': client,
        'iat': iat,
        'exp': exp,
        'scope': scopes,
    }
    return jwt.encode(header, payload, private_key).decode('ascii')


def query_client(path: pathlib.Path, client: str, is_name: bool = False) -> Client:
    if is_name:
        keymap_path = get_keymap_path(path)
        if keymap_path.exists():
            data = yaml.load(keymap_path)
            if client not in data.keys():
                raise (InvalidClientError(description='Invalid client name'))
            client = data[client]
    client_file = get_client_file_path(path, client)
    try:
        data = yaml.load(client_file)
    except FileNotFoundError:
        raise (InvalidClientError(description='Invalid client id or secret'))
    if not isinstance(data['scopes'], list):
        raise Exception(f'Client {client_file} scopes must be list of scopes.')
    client_id = data["client_id"]
    client_name = data["client_name"] if ("client_name" in data.keys() and data["client_name"]) else None
    client = Client(id_=client_id, name_=client_name, secret_hash=data['client_secret_hash'],
                    scopes=data['scopes'])
    return client


def get_client_file_path(path: pathlib.Path, client: str) -> pathlib.Path:
    is_uuid = is_str_uuid(client)
    client_file = path / 'unknown' / f'{client}.yml'
    if is_uuid:
        client_file = path / 'id' / client[:2] / client[2:4] / f'{client[4:]}.yml'
    return client_file


def check_scope(context: Context, scope: str):
    config = context.get('config')
    token = context.get('auth.token')
    token.check_scope(f'{config.scope_prefix}{scope}')


def get_scope_name(
    context: Context,
    node: Union[Namespace, Model, Property],
    action: Action,
) -> str:
    config = context.get('config')

    if isinstance(node, Namespace):
        name = node.name
    elif isinstance(node, Model):
        name = node.model_type()
    elif isinstance(node, Property):
        name = node.model.model_type() + '_' + node.place
    else:
        raise Exception(f"Unknown node type {node}.")

    return name_to_scope(
        '{prefix}{name}_{action}' if name else '{prefix}{action}',
        name,
        maxlen=config.scope_max_length,
        params={
            'prefix': config.scope_prefix,
            'action': action.value,
        },
    )


def get_clients_list(path: pathlib.Path) -> list:
    items = os.listdir(path)
    ids = []
    if 'id' in items:
        id_items = os.listdir(path / 'id')
        for id0 in id_items:
            if len(id0) == 2:
                id0_items = os.listdir(path / 'id' / id0)
                for id1 in id0_items:
                    if len(id1) == 2:
                        id1_items = os.listdir(path / 'id' / id0 / id1)
                        for uuid_item in id1_items:
                            if uuid_item.endswith('.yml') and len(uuid_item) == 36:
                                ids.append(f'{id0}{id1}{uuid_item[:-4]}')
    return ids


def authorized(
    context: Context,
    node: Union[Namespace, Model, Property],
    action: Action,
    *,
    throw: bool = False,
    scope_formatter: ScopeFormatterFunc = None,
):
    config: Config = context.get('config')
    token = context.get('auth.token')

    # Unauthorized clients can only access open nodes.
    unauthorized = token.get_client_id() == get_client_id_from_name(get_clients_path(config), config.default_auth_client)

    open_node = node.access >= Access.open
    if unauthorized and not open_node:
        if throw:
            raise AuthorizedClientsOnly()
        else:
            return False

    # Private nodes can only be accessed with explicit node scope.
    scopes = [node]

    # Protected and higher level nodes can be accessed with parent nodes scopes.
    if node.access > Access.private:
        ns = None

        if isinstance(node, Property):
            # Hidden nodes also require explicit scope.
            # XXX: `hidden` parameter should only be used for API control, not
            #      access control. See docs.
            if not node.hidden:
                scopes.append(node.model)
                scopes.append(node.model.ns)
                ns = node.model.ns
        elif isinstance(node, Model):
            scopes.append(node.ns)
            ns = node.ns
        elif isinstance(node, Namespace):
            ns = node

        # Add all parent namespace scopes too.
        if ns:
            scopes.extend(ns.parents())

    # Build scope names.
    scope_formatter = scope_formatter or config.scope_formatter
    if not isinstance(action, (list, tuple)):
        action = [action]
    scopes = [
        scope_formatter(context, scope, act)
        for act in action
        for scope in scopes
    ]

    # Check if client has at least one of required scopes.
    if throw:
        token.check_scope(scopes, operator='OR')
    else:
        return token.valid_scope(scopes, operator='OR')


def auth_server_keys_exists(path: pathlib.Path):
    return (
        (path / 'keys/private.json').exists() and
        (path / 'keys/public.json').exists()
    )


def gen_auth_server_keys(
    path: pathlib.Path,
    *,
    overwrite: bool = False,
    exist_ok: bool = False,
) -> Tuple[pathlib.Path, pathlib.Path]:
    path = path / 'keys'
    path.mkdir(exist_ok=True)

    files = (
        path / 'private.json',
        path / 'public.json',
    )

    if overwrite:
        create = True
    else:
        create = False
        for file in files:
            if file.exists():
                if not exist_ok:
                    raise KeyFileExists(f"{file} file already exists.")
            else:
                create = True

    if create:
        key = create_key_pair()
        keys = (key, key.public_key())
        for k, file in zip(keys, files):
            with file.open('w') as f:
                json.dump(jwk.dumps(k), f, indent=4, ensure_ascii=False)

    return files


class KeyFileExists(Exception):
    pass


def client_exists(path: pathlib.Path, client: str) -> bool:
    client_file = get_client_file_path(path, client)
    if client_file.exists():
        return True
    return False


def client_name_exists(path: pathlib.Path, client: str) -> bool:
    keymap_path = get_keymap_path(path)
    if keymap_path.exists():
        keymap = yaml.load(keymap_path)
        if client in keymap.keys():
            return True
    return False


def create_client_file(
    path: pathlib.Path,
    name: str,
    client_id: str,
    secret: str = None,
    scopes: List[str] = None,
    *,
    add_secret: bool = False,
) -> Tuple[pathlib.Path, dict]:
    client_file = get_client_file_path(path, client_id)
    keymap_path = get_keymap_path(path)

    if client_file.exists():
        raise ClientAlreadyExists(client_id=client_file)

    keymap = {}
    if keymap_path.exists():
        keymap = yaml.load(keymap_path)

    if client_name_exists(path, name):
        raise ClientWithNameAlreadyExists(client_name=name)

    os.makedirs(path / "helpers", exist_ok=True)
    os.makedirs(path / "id" / client_id[:2] / client_id[2:4], exist_ok=True)

    secret = secret or passwords.gensecret(32)
    secret_hash = passwords.crypt(secret)

    data = write = {
        'client_id': client_id,
        'client_name': name,
        'client_secret': secret,
        'client_secret_hash': secret_hash,
        'scopes': scopes or [],
    }
    keymap[name] = client_id

    if not add_secret:
        write = data.copy()
        del write['client_secret']
    yml.dump(write, client_file)
    yml.dump(keymap, keymap_path)

    return client_file, data


class ClientExist(Exception):
    pass


def delete_client_file(path: pathlib.Path, client_id: str):
    if client_exists(path, client_id):
        keymap_path = get_keymap_path(path)
        remove_path = get_client_file_path(path, client_id)
        if keymap_path.exists():
            keymap = yaml.load(keymap_path)
            changed = False
            keymap_values = list(keymap.values())
            keymap_keys = list(keymap.keys())
            if client_id in keymap_values:
                del keymap[keymap_keys[keymap_values.index(client_id)]]
                changed = True

            if changed:
                yml.dump(keymap, keymap_path)
        os.remove(remove_path)
        try:
            level_1 = remove_path.parent
            level_2 = remove_path.parent.parent
            level_1.rmdir()
            level_2.rmdir()
        except Exception:
            '''
                Remove only empty folders
            '''
    else:
        raise (InvalidClientError(description='Invalid client id or secret'))


def update_client_file(
    context: Context,
    path: pathlib.Path,
    client_id: str,
    name: str,
    secret: str,
    scopes: list
):
    if client_exists(path, client_id):
        config = context.get("config")
        client = query_client(get_clients_path(config), client_id)
        keymap_path = get_keymap_path(path)

        new_name = name if name else client.name
        new_secret_hash = passwords.crypt(secret) if secret else client.secret_hash
        new_scopes = scopes if scopes is not None else client.scopes

        client_path = get_client_file_path(path, client_id)
        keymap = {}
        if keymap_path.exists():
            keymap = yaml.load(keymap_path)
        if new_name != client.name:
            if client_name_exists(path, new_name):
                raise ClientWithNameAlreadyExists(client_name=new_name)

        new_data = {
            "client_id": client.id,
            "client_name": new_name,
            "client_secret_hash": new_secret_hash,
            "scopes": list(new_scopes)
        }

        yml.dump(new_data, client_path)
        if keymap:
            changed = False
            if client_id in keymap.keys():
                if client_id != new_name:
                    del keymap[client_id]
                    keymap[new_name] = client.id
                    changed = True

            if changed:
                yml.dump(keymap, keymap_path)
        return new_data
    else:
        raise (InvalidClientError(description='Invalid client id or secret'))


def migrate_old_client_file(path: pathlib.Path, client_file: str):
    try:
        old_data = yaml.load(path / client_file)
    except FileNotFoundError:
        raise (InvalidClientError(description='Invalid client id or secret'))

    old_path = path / client_file

    old_data["client_name"] = old_data["client_id"]
    if not is_str_uuid(old_data["client_id"]):
        old_data["client_id"] = str(uuid.uuid4())
    new_client_file = old_data["client_id"]

    new_path = path / 'id' / new_client_file[:2] / new_client_file[2:4] / f'{new_client_file[4:]}.yml'

    os.makedirs(path / 'id' / new_client_file[:2] / new_client_file[2:4], exist_ok=True)
    shutil.copyfile(old_path, new_path)

    with open(new_path) as fp:
        data = yaml.load(fp)
    data["client_id"] = old_data["client_id"]
    data["client_name"] = old_data["client_name"]
    yml.dump({
        "client_id": data["client_id"],
        "client_name": data["client_name"],
        "client_secret_hash": data["client_secret_hash"],
        "scopes": data["scopes"]
    }, new_path)


def create_client_file_name_id_mapping(path: pathlib.Path):
    os.makedirs(path / 'helpers', exist_ok=True)
    keymap_path = get_keymap_path(path)

    items = os.listdir(path)
    keymap = {}
    if 'id' in items:
        id_items = os.listdir(path / 'id')
        for id0 in id_items:
            if len(id0) == 2:
                id0_items = os.listdir(path / 'id' / id0)
                for id1 in id0_items:
                    if len(id1) == 2:
                        id1_items = os.listdir(path / 'id' / id0 / id1)
                        for uuid_item in id1_items:
                            if uuid_item.endswith('.yml') and len(uuid_item) == 36:
                                try:
                                    data = yaml.load(path / 'id' / id0 / id1 / uuid_item)
                                    keymap[data["client_name"]] = data["client_id"]
                                except FileNotFoundError:
                                    raise (InvalidClientError(description='Invalid client id or secret'))
    yml.dump(keymap, keymap_path)


def handle_auth_client_files(config: Config):
    path = get_clients_path(config)
    if path.exists():
        if not (path / 'id').exists():
            items = os.listdir(path)
            if items:
                echo("Client file migrations has started")
                for item in items:
                    if item.endswith('.yml'):
                        migrate_old_client_file(path, item)

        create_client_file_name_id_mapping(path)


def get_client_id_from_name(path: pathlib.Path, client_name: str):
    keymap_path = get_keymap_path(path)
    if keymap_path.exists():
        keymap = yaml.load(keymap_path)
        if client_name in keymap.keys():
            return keymap[client_name]
    return None


def get_keymap_path(path: pathlib.Path):
    return path / 'helpers' / 'keymap.yml'


def get_clients_path(path: Union[Config, pathlib.Path]):
    if isinstance(path, pathlib.Path):
        return path / 'clients'
    return path.config_path / 'clients'




from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.memory.components import MemoryManifest


@commands.load.register(Context, MemoryManifest)
def load(
     context: Context,
     manifest: MemoryManifest,
     *,
     into: Manifest = None,
     freezed: bool = True,
     rename_duplicates: bool = False,
     load_internal: bool = True,
):
    if load_internal:
        target = into or manifest
        if '_schema' not in target.models:
            store = context.get('store')
            commands.load(context, store.internal, into=target)

    for source in manifest.sync:
        commands.load(
            context, source,
            into=into or manifest,
            freezed=freezed,
            rename_duplicates=rename_duplicates,
            load_internal=load_internal,
        )



from typing import Optional

from spinta import commands
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.sql.components import SqlManifest


@commands.configure.register(Context, SqlManifest)
def configure(context: Context, manifest: SqlManifest):
    rc: RawConfig = context.get('rc')
    path: Optional[str] = rc.get('manifests', manifest.name, 'path')
    prepare: Optional[str] = rc.get('manifests', manifest.name, 'prepare')
    manifest.prepare = prepare
    manifest.path = path
